/* SPDX-License-Identifier: GPL-2.0 */
/* Copyright (C) 2009  PaX Team <pageexec@freemail.hu> */

#ifndef _ASM_X86_ALTERNATIVE_H
#define _ASM_X86_ALTERNATIVE_H

#include <linux/types.h>
#include <linux/stringify.h>
#include <asm/asm.h>
#include <asm/irq_vectors.h>

#define ALTINSTR_FLAG_INV	(1 << 15)
#define ALT_NOT(feat)		((feat) | ALTINSTR_FLAG_INV)

#ifndef __ASSEMBLY__

#include <linux/stddef.h>

/*
 * Alternative inline assembly for SMP.
 *
 * The LOCK_PREFIX macro defined here replaces the LOCK and
 * LOCK_PREFIX macros used everywhere in the source tree.
 *
 * SMP alternatives use the same data structures as the other
 * alternatives and the X86_FEATURE_UP flag to indicate the case of a
 * UP system running a SMP kernel.  The existing apply_alternatives()
 * works fine for patching a SMP kernel for UP.
 *
 * The SMP alternative tables can be kept after boot and contain both
 * UP and SMP versions of the instructions to allow switching back to
 * SMP at runtime, when hotplugging in a new CPU, which is especially
 * useful in virtualized environments.
 *
 * The very common lock prefix is handled as special case in a
 * separate table which is a pure address list without replacement ptr
 * and size information.  That keeps the table sizes small.
 */

#ifdef CONFIG_SMP
#define LOCK_PREFIX_HERE \
		".pushsection .smp_locks,\"a\"\n"	\
		".balign 4\n"				\
		".long 671f - .\n" /* offset */		\
		".popsection\n"				\
		"671:"

#define LOCK_PREFIX LOCK_PREFIX_HERE "\n\tlock; "

#else /* ! CONFIG_SMP */
#define LOCK_PREFIX_HERE ""
#define LOCK_PREFIX ""
#endif

/*
 * objtool annotation to ignore the alternatives and only consider the original
 * instruction(s).
 */
#define ANNOTATE_IGNORE_ALTERNATIVE				\
	"999:\n\t"						\
	".pushsection .discard.ignore_alts\n\t"			\
	".long 999b - .\n\t"					\
	".popsection\n\t"

struct alt_instr {
	s32 instr_offset;	/* original instruction */
	s32 repl_offset;	/* offset to replacement instruction */
	u16 cpuid;		/* cpuid bit set for replacement */
	u8  instrlen;		/* length of original instruction */
	u8  replacementlen;	/* length of new instruction */
} __packed;

/*
 * Debug flag that can be tested to see whether alternative
 * instructions were patched in already:
 */
extern int alternatives_patched;

extern void alternative_instructions(void);
extern void apply_alternatives(struct alt_instr *start, struct alt_instr *end, const u8 *altinstr_replacements, const u8 *altinstr_replacements_end);
extern void apply_returns(s32 *start, s32 *end);

struct module;

#ifdef CONFIG_SMP
extern void alternatives_smp_module_add(struct module *mod, char *name,
					void *locks, void *locks_end,
					void *text, void *text_end);
extern void alternatives_smp_module_del(struct module *mod);
extern void alternatives_enable_smp(void);
extern int alternatives_text_reserved(void *start, void *end);
extern bool skip_smp_alternatives;
#else
static inline void alternatives_smp_module_add(struct module *mod, char *name,
					       void *locks, void *locks_end,
					       void *text, void *text_end) {}
static inline void alternatives_smp_module_del(struct module *mod) {}
static inline void alternatives_enable_smp(void) {}
static inline int alternatives_text_reserved(void *start, void *end)
{
	return 0;
}
#endif	/* CONFIG_SMP */

/* can only use gas macros in C when GRKERNSEC and plugins are enabled */
#if defined(CONFIG_PLUGIN_WANTS_ASMMACRO)
#define PAX_RET_NOSPEC	"pax_ret_nospec"
#else
#define PAX_RET_NOSPEC	"ret"
#endif

#ifdef CONFIG_SLS
#define PAX_DIRECT_SLS_CALL(target)	"pax_direct_sls_call " target
#define PAX_DIRECT_SLS_CALL_HASH(target, hash) "pax_direct_sls_call " target ", " hash
#define PAX_INDIRECT_SLS_CALL_HASH(target, hash) "pax_indirect_call " target ", " hash
#define PAX_SLS_CALL(target)		"pax_sls_call " target
#define PAX_JMP(target)			"pax_jmp " target
#else
#define PAX_DIRECT_SLS_CALL(target)	PAX_DIRECT_CALL(target)
#define PAX_DIRECT_SLS_CALL_HASH(target, hash) PAX_DIRECT_CALL_HASH(target, hash)
#define PAX_INDIRECT_SLS_CALL_HASH(target, hash) ANNOTATE_RETPOLINE_SAFE "call *" target
#define PAX_SLS_CALL(target)		"call " target
#define PAX_JMP(target)			"jmp " target
#endif


#ifdef CONFIG_PAX_RAP_HASH
#define PAX_DIRECT_CALL(target) "pax_direct_call " target
#define PAX_DIRECT_CALL_HASH(target, hash) "pax_direct_call " target ", " hash
#define PAX_DIRECT_CALL_HASHVAL(target, hashval) "pax_direct_call " target ", hashval=" hashval
#define PAX_PARAVIRT_CALL(target, hash) "pax_paravirt_call " target ", " hash
#define PAX_INDIRECT_CALL(target, extra) CALL_NOSPEC_INLINE(target, extra)
#define PAX_RET(extra) "pax_ret " extra
#else
#define PAX_DIRECT_CALL(target) "call " target
#define PAX_DIRECT_CALL_HASH(target, hash) "call " target
#define PAX_DIRECT_CALL_HASHVAL(target, hashval) "call " target
#define PAX_PARAVIRT_CALL(target, hash) PAX_INDIRECT_SLS_CALL_HASH(target, hash)
#define PAX_INDIRECT_CALL(target, extra) CALL_NOSPEC
#define PAX_RET(extra) PAX_RET_NOSPEC
#endif

#define b_replacement(num)	"664"#num
#define e_replacement(num)	"665"#num

#define alt_end_marker		"663"
#define alt_slen		"662b-661b"
#define alt_total_slen		alt_end_marker"b-661b"
#define alt_rlen(num)		e_replacement(num)"f-"b_replacement(num)"f"

#define OLDINSTR(oldinstr, num)						\
	"# ALT: oldnstr\n"						\
	"661:\n\t" oldinstr "\n662:\n"					\
	"# ALT: padding\n"						\
	".skip -(((" alt_rlen(num) ")-(" alt_slen ")) > 0) * "		\
		"((" alt_rlen(num) ")-(" alt_slen ")),0x90\n"		\
	alt_end_marker ":\n"

/*
 * gas compatible max based on the idea from:
 * http://graphics.stanford.edu/~seander/bithacks.html#IntegerMinOrMax
 *
 * The additional "-" is needed because gas uses a "true" value of -1.
 */
#define alt_max_short(a, b)	"((" a ") ^ (((" a ") ^ (" b ")) & -(-((" a ") < (" b ")))))"

/*
 * Pad the second replacement alternative with additional NOPs if it is
 * additionally longer than the first replacement alternative.
 */
#define OLDINSTR_2(oldinstr, num1, num2) \
	"# ALT: oldinstr2\n"									\
	"661:\n\t" oldinstr "\n662:\n"								\
	"# ALT: padding2\n"									\
	".skip -((" alt_max_short(alt_rlen(num1), alt_rlen(num2)) " - (" alt_slen ")) > 0) * "	\
		"(" alt_max_short(alt_rlen(num1), alt_rlen(num2)) " - (" alt_slen ")), 0x90\n"	\
	alt_end_marker ":\n"

#define OLDINSTR_3(oldinsn, n1, n2, n3)								\
	"# ALT: oldinstr3\n"									\
	"661:\n\t" oldinsn "\n662:\n"								\
	"# ALT: padding3\n"									\
	".skip -((" alt_max_short(alt_max_short(alt_rlen(n1), alt_rlen(n2)), alt_rlen(n3))	\
		" - (" alt_slen ")) > 0) * "							\
		"(" alt_max_short(alt_max_short(alt_rlen(n1), alt_rlen(n2)), alt_rlen(n3))	\
		" - (" alt_slen ")), 0x90\n"							\
	alt_end_marker ":\n"

#define ALTINSTR_ENTRY(feature, num)					      \
	" .long 661b - .\n"				/* label           */ \
	" .long " b_replacement(num)"f - .\n"		/* new instruction */ \
	" .word " __stringify(feature) "\n"		/* feature bit     */ \
	" .byte " alt_total_slen "\n"			/* source len      */ \
	" .byte " alt_rlen(num) "\n"			/* replacement len */

#define ALTINSTR_REPLACEMENT(newinstr, num)		/* replacement */	\
	"# ALT: replacement " #num "\n"						\
	b_replacement(num)":\n\t" newinstr "\n" e_replacement(num) ":\n"

/* alternative assembly primitive: */
#define ALTERNATIVE(oldinstr, newinstr, feature)			\
	OLDINSTR(oldinstr, 1)						\
	".pushsection .altinstructions,\"a\"\n"				\
	ALTINSTR_ENTRY(feature, 1)					\
	".popsection\n"							\
	".pushsection .altinstr_replacement, \"ax\"\n"			\
	ALTINSTR_REPLACEMENT(newinstr, 1)				\
	".popsection\n"

#define ALTERNATIVE_2(oldinstr, newinstr1, feature1, newinstr2, feature2)\
	OLDINSTR_2(oldinstr, 1, 2)					\
	".pushsection .altinstructions,\"a\"\n"				\
	ALTINSTR_ENTRY(feature1, 1)					\
	ALTINSTR_ENTRY(feature2, 2)					\
	".popsection\n"							\
	".pushsection .altinstr_replacement, \"ax\"\n"			\
	ALTINSTR_REPLACEMENT(newinstr1, 1)				\
	ALTINSTR_REPLACEMENT(newinstr2, 2)				\
	".popsection\n"

/* If @feature is set, patch in @newinstr_yes, otherwise @newinstr_no. */
#define ALTERNATIVE_TERNARY(oldinstr, feature, newinstr_yes, newinstr_no) \
	ALTERNATIVE_2(oldinstr, newinstr_no, X86_FEATURE_ALWAYS,	\
		      newinstr_yes, feature)

#define ALTERNATIVE_3(oldinsn, newinsn1, feat1, newinsn2, feat2, newinsn3, feat3) \
	OLDINSTR_3(oldinsn, 1, 2, 3)						\
	".pushsection .altinstructions,\"a\"\n"					\
	ALTINSTR_ENTRY(feat1, 1)						\
	ALTINSTR_ENTRY(feat2, 2)						\
	ALTINSTR_ENTRY(feat3, 3)						\
	".popsection\n"								\
	".pushsection .altinstr_replacement, \"ax\"\n"				\
	ALTINSTR_REPLACEMENT(newinsn1, 1)					\
	ALTINSTR_REPLACEMENT(newinsn2, 2)					\
	ALTINSTR_REPLACEMENT(newinsn3, 3)					\
	".popsection\n"

/*
 * Alternative instructions for different CPU types or capabilities.
 *
 * This allows to use optimized instructions even on generic binary
 * kernels.
 *
 * length of oldinstr must be longer or equal the length of newinstr
 * It can be padded with nops as needed.
 *
 * For non barrier like inlines please define new variants
 * without volatile and memory clobber.
 */
#define alternative(oldinstr, newinstr, feature)			\
	asm_inline volatile (ALTERNATIVE(oldinstr, newinstr, feature) : : : "memory")

#define alternative_2(oldinstr, newinstr1, feature1, newinstr2, feature2) \
	asm_inline volatile(ALTERNATIVE_2(oldinstr, newinstr1, feature1, newinstr2, feature2) ::: "memory")

#define alternative_ternary(oldinstr, feature, newinstr_yes, newinstr_no) \
	asm_inline volatile(ALTERNATIVE_TERNARY(oldinstr, feature, newinstr_yes, newinstr_no) ::: "memory")

/*
 * Alternative inline assembly with input.
 *
 * Peculiarities:
 * No memory clobber here.
 * Argument numbers start with 1.
 * Leaving an unused argument 0 to keep API compatibility.
 */
#define alternative_input(oldinstr, newinstr, feature, input...)	\
	asm_inline volatile (ALTERNATIVE(oldinstr, newinstr, feature)	\
		: : "i" (0), ## input)

/*
 * This is similar to alternative_input. But it has two features and
 * respective instructions.
 *
 * If CPU has feature2, newinstr2 is used.
 * Otherwise, if CPU has feature1, newinstr1 is used.
 * Otherwise, oldinstr is used.
 */
#define alternative_input_2(oldinstr, newinstr1, feature1, newinstr2,	     \
			   feature2, input...)				     \
	asm_inline volatile(ALTERNATIVE_2(oldinstr, newinstr1, feature1,     \
		newinstr2, feature2)					     \
		: : "i" (0), ## input)

/* Like alternative_input, but with a single output argument */
#define alternative_io(oldinstr, newinstr, feature, output, input...)	\
	asm_inline volatile (ALTERNATIVE(oldinstr, newinstr, feature)	\
		: output : "i" (0), ## input)

/* Like alternative_io, but for replacing a direct call with another one. */
#define alternative_call(oldfunc, newfunc, feature, output, input...)	\
	asm_inline volatile (ALTERNATIVE(PAX_DIRECT_SLS_CALL("%P[old]"), PAX_DIRECT_SLS_CALL("%P[new]"), feature) \
		: output, ASM_CALL_CONSTRAINT : [old] "i" (oldfunc), [new] "i" (newfunc), ## input)

/*
 * Like alternative_call, but there are two features and respective functions.
 * If CPU has feature2, function2 is used.
 * Otherwise, if CPU has feature1, function1 is used.
 * Otherwise, old function is used.
 */
#define alternative_call_2(oldfunc, newfunc1, feature1, newfunc2, feature2,   \
			   output, input...)				      \
	asm_inline volatile (ALTERNATIVE_2(PAX_DIRECT_SLS_CALL("%P[old]"), PAX_DIRECT_SLS_CALL("%P[new1]"), feature1,\
		PAX_DIRECT_SLS_CALL("%P[new2]"), feature2)			      \
		: output, ASM_CALL_CONSTRAINT				      \
		: [old] "i" (oldfunc), [new1] "i" (newfunc1),		      \
		  [new2] "i" (newfunc2), ## input)

/*
 * use this macro(s) if you need more than one output parameter
 * in alternative_io
 */
#define ASM_OUTPUT2(a...) a

/*
 * use this macro if you need clobbers but no inputs in
 * alternative_{input,io,call}()
 */
#define ASM_NO_INPUT_CLOBBER(clbr...) "i" (0) : clbr

#ifdef CONFIG_PAX_REFCOUNT
#define __PAX_REFCOUNT(size, cc)			\
	"j"cc" 111f\n"					\
	".if "__stringify(size)" == 4\n\t"		\
	".pushsection .text.refcount_overflow."cc"\n"	\
	".elseif "__stringify(size)" == -4\n\t"		\
	".pushsection .text.refcount_underflow."cc"\n"	\
	".elseif "__stringify(size)" == 8\n\t"		\
	".pushsection .text.refcount64_overflow."cc"\n"	\
	".elseif "__stringify(size)" == -8\n\t"		\
	".pushsection .text.refcount64_underflow."cc"\n"\
	".else\n"					\
	".error \"invalid size\"\n"			\
	".endif\n"					\
	"111:\tlea %[counter],%%"_ASM_CX"\n\t"		\
	"222:\tint $"__stringify(X86_REFCOUNT_VECTOR)"\n"\
	".popsection\n"					\
	"333:\n"					\
	_ASM_EXTABLE(222b, 333b)

#define PAX_REFCOUNT_OVERFLOW(size)	__PAX_REFCOUNT(size, "o")
#define PAX_REFCOUNT_UNDERFLOW(size)	__PAX_REFCOUNT(-(size), "o")
#else
#define __PAX_REFCOUNT(size, cc)
#define PAX_REFCOUNT_OVERFLOW(size)
#define PAX_REFCOUNT_UNDERFLOW(size)
#endif

#else /* __ASSEMBLY__ */

#include <linux/linkage.h>
#include <asm/page.h>
#include <asm/percpu.h>

#ifdef CONFIG_SMP
	.macro LOCK_PREFIX
.Llock_prefix_\@:
	lock
	.pushsection .smp_locks,"a"
	.balign 4
	.long .Llock_prefix_\@ - .
	.popsection
	.endm
#else
	.macro LOCK_PREFIX
	.endm
#endif

/*
 * objtool annotation to ignore the alternatives and only consider the original
 * instruction(s).
 */
.macro ANNOTATE_IGNORE_ALTERNATIVE
	.Lannotate_\@:
	.pushsection .discard.ignore_alts
	.long .Lannotate_\@ - .
	.popsection
.endm

.macro pax_force_retaddr_bts rip=0
#ifdef KERNEXEC_PLUGIN
	btsq $63,\rip(%rsp)
#endif
.endm

#if defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS) && defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR)
#error PAX: the KERNEXEC BTS and OR methods must not be enabled at once
#endif

#if defined(CONFIG_PAX_RAP_XOR) && defined(CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR)
#error PAX: RAP and the KERNEXEC OR methods must not be enabled at once
#endif

.macro pax_force_retaddr rip=0
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS
	btsq $63,\rip(%rsp)
#endif
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	orq %r12,\rip(%rsp)
#endif
.endm

.macro pax_force_fptr ptr:req
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_BTS
	btsq $63,\ptr
#endif
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	orq %r12,\ptr
#endif
.endm

.macro pax_set_fptr_mask altstack=0
#ifdef CONFIG_PAX_KERNEXEC_PLUGIN_METHOD_OR
	movabs $0x8000000000000000,%r12
#endif
#ifdef CONFIG_PAX_RAP_XOR
	.if \altstack
	movq PER_CPU_VAR(cpu_current_top_of_stack),%r12
	.else
	movq %rsp,%r12
	.endif
	andq $~(THREAD_SIZE_asm-1),%r12
	movq (%r12),%r12
#endif
.endm

.macro pax_jmp target:req
	jmp \target
#ifdef CONFIG_SLS
	int3
#endif
.endm

#ifdef CONFIG_PAX_RAP_HASH
.macro misalign align:req offset:req pad=0xcc
#ifdef HAVE_MISALIGN
	.set .Loffset_\@,\offset
	.org .+(\align+(.Loffset_\@&(\align-1))-(.%\align))&(\align-1),\pad
#endif
.endm

.macro rap_retloc callee="" hashval=0 call_size=0
	.set .Lcall_size_\@,\call_size
	.set .Lrap_retloc_len_\@,.Lrap_retloc_end_\@-.Lrap_retloc_start_\@
//	.skip (16 - ((.Lrap_retloc_end_\@ - .Lrap_retloc_start_\@ - \call_size) % 16)) % 16, 0xcc
	misalign align=8 offset=-.Lrap_retloc_len_\@
.Lrap_retloc_start_\@:
	.ifnb \callee
	ASM_RAP_RET_HASH(__rap_hash_ret_\callee)
	.elseif \hashval
	ASM_RAP_RET_HASH(\hashval)
	.else
	.error "either 'callee' or 'hashval' must be specified"
	.endif
	.skip RAP_RET_HASH_OFFSET-RAP_HASH_LEN-.Lcall_size_\@,0xcc
.Lrap_retloc_end_\@:
.endm

.macro rap_call target:req hashsym="" hashval=0 sym="" lfence=0 annotation=
//	.p2align 5,,2
	pax_jmp .Lrap_call_target_\@
	.ifnb \hashsym
	rap_retloc callee=\hashsym call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.elseif \hashval
	rap_retloc hashval=\hashval call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.else
	rap_retloc callee=\target call_size=.Lrap_call_target_end_\@-.Lrap_call_target_\@
	.endif

	.ifnb \sym
SYM_INNER_LABEL(\sym, SYM_L_GLOBAL)
	.endif

.Lrap_call_target_\@:
	.if \lfence
	lfence
	.endif
	\annotation
	call \target
.Lrap_call_target_end_\@:
.endm

.macro rap_ret func:req
	RET
.endm
#endif

.macro pax_indirect_jmp target:req
	pax_jmp *\target
.endm

.macro pax_direct_call_global target:req hashsym:req
#ifdef CONFIG_SLS
	xor %eax, %eax
#endif
#ifdef CONFIG_PAX_RAP_HASH
	rap_call \target, hashsym=\hashsym, sym=\hashsym
#else
SYM_INNER_LABEL(\hashsym, SYM_L_GLOBAL)
	call \target
#endif
.endm

.macro pax_indirect_call target:req hashsym:req lfence=0
#ifdef CONFIG_SLS
	.ifc \target,%_ASM_AX
	.else
	xor %eax, %eax
	.endif
#endif
#ifdef CONFIG_PAX_RAP_HASH
	rap_call *\target, hashsym=\hashsym, annotation=ANNOTATE_RETPOLINE_SAFE, lfence=\lfence
#else
	.if \lfence
	lfence
	.endif
	ANNOTATE_RETPOLINE_SAFE
	call *\target
#endif
.endm

.macro pax_direct_call target:req hashsym="" hashval=0
#ifdef CONFIG_PAX_RAP_HASH
	rap_call \target, hashsym=\hashsym, hashval=\hashval
#else
	call \target
#endif
.endm

.macro pax_direct_call_intra target:req hashsym="" hashval=0
#ifdef CONFIG_PAX_RAP_HASH
	rap_call \target, hashsym=\hashsym, hashval=\hashval, annotation=ANNOTATE_INTRA_FUNCTION_CALL
#else
	call \target
#endif
.endm

.macro pax_paravirt_call target:req hashsym:req
	pax_indirect_call \target, \hashsym
.endm

.macro pax_direct_sls_call target:req hashsym="" hashval=0
#ifdef CONFIG_SLS
	xor %eax, %eax
#endif
	pax_direct_call \target, hashsym=\hashsym, hashval=\hashval
.endm

.macro pax_sls_call target:req
#ifdef CONFIG_SLS
	xor %eax, %eax
#endif
	call \target
.endm

.macro pax_retloc callee:req sym=""
#ifdef CONFIG_PAX_RAP_HASH
	.ifnb \sym
	rap_hash.\sym:
	.endif
	rap_retloc callee=\callee
	.ifnb \sym
	.type rap_hash.\sym, @function
	.size rap_hash.\sym, . - rap_hash.\sym
	.endif
#else
#endif
.endm

.macro pax_ret func:req
	pax_force_retaddr
#ifdef CONFIG_PAX_RAP_HASH
	rap_ret \func
#else
	RET
#endif
.endm

/* expands to pax_jmp under RETHUNK, and will add the appropriate SLS barrier */
.macro pax_ret_nospec
	RET
.endm

.macro pax_ret_nospec_alternative
#ifdef CONFIG_RETHUNK
	pax_jmp __x86_return_thunk
#else
	RET
#endif
.endm

/*
 * Issue one struct alt_instr descriptor entry (need to put it into
 * the section .altinstructions, see below). This entry contains
 * enough information for the alternatives patching code to patch an
 * instruction. See apply_alternatives().
 */
.macro altinstruction_entry orig alt feature orig_len alt_len
	.long \orig - .
	.long \alt - .
	.word \feature
	.byte \orig_len
	.byte \alt_len
.endm

/*
 * Define an alternative between two instructions. If @feature is
 * present, early code in apply_alternatives() replaces @oldinstr with
 * @newinstr. ".skip" directive takes care of proper instruction padding
 * in case @newinstr is longer than @oldinstr.
 */
.macro ALTERNATIVE oldinstr, newinstr, feature
# define old_len		.Lold_pad_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_\@
# define new_len		.Lnew_end_\@-.Lnew_\@
.Lold_\@\():
	\oldinstr
.Lold_pad_\@\():
	.skip -(((new_len)-(old_len)) > 0) * ((new_len)-(old_len)),0x90
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_\@,.Lnew_\@,\feature,old_full_len,new_len
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew_\@\():
	\newinstr
.Lnew_end_\@\():
	.popsection
# undef old_len
# undef old_full_len
# undef new_len
.endm

/*
 * gas compatible max based on the idea from:
 * http://graphics.stanford.edu/~seander/bithacks.html#IntegerMinOrMax
 *
 * The additional "-" is needed because gas uses a "true" value of -1.
 */
#define alt_max_short(a, b)	((a) ^ (((a) ^ (b)) & -(-((a) < (b)))))


/*
 * Same as ALTERNATIVE macro above but for two alternatives. If CPU
 * has @feature1, it replaces @oldinstr with @newinstr1. If CPU has
 * @feature2, it replaces @oldinstr with @feature2.
 */
.macro ALTERNATIVE_2 oldinstr, newinstr1, feature1, newinstr2, feature2
# define old_len		.Lold_pad_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_\@
# define new1_len		.Lnew1_end_\@-.Lnew1_\@
# define new2_len		.Lnew2_end_\@-.Lnew2_\@
.Lold_\@\():
	\oldinstr
.Lold_pad_\@\():
	.skip -((alt_max_short(new1_len, new2_len) - (old_len)) > 0) * \
		(alt_max_short(new1_len, new2_len) - (old_len)),0x90
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_\@,.Lnew1_\@,\feature1,old_full_len,new1_len
	altinstruction_entry .Lold_\@,.Lnew2_\@,\feature2,old_full_len,new2_len
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew1_\@\():
	\newinstr1
.Lnew1_end_\@\():

.Lnew2_\@\():
	\newinstr2
.Lnew2_end_\@:
	.popsection
# undef old_len
# undef old_full_len
# undef new1_len
# undef new2_len
.endm

.macro ALTERNATIVE_2_PREPAD oldinstr, newinstr1, feature1, newinstr2, feature2
# define old_len		.Lold_end_\@-.Lold_\@
# define old_full_len		.Lold_end_\@-.Lold_pad_\@
# define new1_len		.Lnew1_end_\@-.Lnew1_\@
# define new1_full_len		.Lnew1_end_\@-.Lnew1_pad_\@
# define new2_len		.Lnew2_end_\@-.Lnew2_\@
# define new2_full_len		.Lnew2_end_\@-.Lnew2_pad_\@
.Lold_pad_\@\():
	.skip -((alt_max_short(new1_len, new2_len) - (old_len)) > 0) * \
		(alt_max_short(new1_len, new2_len) - (old_len)),0x90
.Lold_\@\():
	\oldinstr
.Lold_end_\@\():

	.pushsection .altinstructions,"a"
	altinstruction_entry .Lold_pad_\@,.Lnew1_pad_\@,\feature1,old_full_len,new1_full_len
	altinstruction_entry .Lold_pad_\@,.Lnew2_pad_\@,\feature2,old_full_len,new2_full_len
	.popsection

	.pushsection .altinstr_replacement,"ax"
.Lnew1_pad_\@\():
	.skip -((alt_max_short(new2_len, old_len) - (new1_len)) > 0) * \
		(alt_max_short(new2_len, old_len) - (new1_len)),0x90
.Lnew1_\@\():
	\newinstr1
.Lnew1_end_\@\():

.Lnew2_pad_\@\():
	.skip -((alt_max_short(new1_len, old_len) - (new2_len)) > 0) * \
		(alt_max_short(new1_len, old_len) - (new2_len)),0x90
.Lnew2_\@\():
	\newinstr2
.Lnew2_end_\@\():
	.popsection
# undef old_len
# undef old_full_len
# undef new1_len
# undef new1_full_len
# undef new2_len
# undef new2_full_len
.endm

.macro __PAX_REFCOUNT section, counter
#ifdef CONFIG_PAX_REFCOUNT
	jo .Lrefcount_error_\@
	.pushsection .text.\section\().o
.Lrefcount_error_\@:
	lea \counter,%_ASM_CX
.Lrefcount_report_\@:
	int $X86_REFCOUNT_VECTOR
	.popsection
.Lrefcount_continue_\@:
	_ASM_EXTABLE(.Lrefcount_report_\@, .Lrefcount_continue_\@)
#endif
.endm

.macro PAX_REFCOUNT64_OVERFLOW counter
	__PAX_REFCOUNT refcount64_overflow, \counter
.endm

.macro PAX_REFCOUNT64_UNDERFLOW counter
	__PAX_REFCOUNT refcount64_underflow, \counter
.endm

/* If @feature is set, patch in @newinstr_yes, otherwise @newinstr_no. */
#define ALTERNATIVE_TERNARY(oldinstr, feature, newinstr_yes, newinstr_no) \
	ALTERNATIVE_2 oldinstr, newinstr_no, ALT_NOT(feature),	\
	newinstr_yes, feature

#endif /* __ASSEMBLY__ */

#endif /* _ASM_X86_ALTERNATIVE_H */
