/*
 * Copyright 2020-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

GTY(()) tree convert_autostack_decl;
GTY(()) tree free_autostack_decl;

EXPORTED_CONST struct ggc_root_tab gt_ggc_r_gt_autostack[] = {
	INIT_GGC_ROOT_TAB_TREE(convert_autostack_decl),
	INIT_GGC_ROOT_TAB_TREE(free_autostack_decl),
	LAST_GGC_ROOT_TAB
};

static tree autoslab_build_local_var(location_t loc, tree id, tree type, tree ctor, unsigned int align)
{
	tree var;

	var = autoslab_build_var(loc, id, type, ctor);

	SET_DECL_ALIGN(var, align * BITS_PER_UNIT);
	DECL_USER_ALIGN(var) = 1;

	mark_sym_for_renaming(var);
	gimple_add_tmp_var(var);

	return var;
}

static void propagate_single_arg_phis(tree lhs)
{
	bool todo = false;
	gimple_stmt_ptr use_stmt;
	imm_use_iterator imm_iter;

	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		gimple_stmt_iterator gsi;
		gimple_stmt_ptr use_stmt2;
		imm_use_iterator imm_iter2;

		if (!single_arg_phi_p(use_stmt))
			continue;

		FOR_EACH_IMM_USE_STMT(use_stmt2, imm_iter2, gimple_phi_result(use_stmt)) {
			use_operand_p use;

			FOR_EACH_IMM_USE_ON_STMT(use, imm_iter2) {
				SET_USE(use, lhs);
			}
			update_stmt(use_stmt2);
		}

		gsi = gsi_for_stmt(use_stmt);
		remove_phi_node(&gsi, true);
		todo = true;
	}
	if (todo)
		update_ssa(TODO_update_ssa);
}

static bool self_recursive_p(void)
{
	cgraph_node_ptr node;
	cgraph_edge_p edge;

	node = cgraph_get_node(current_function_decl);
	gcc_assert(node);

	for (edge = node->callees; edge; edge = edge->next_callee) {
		bool recursive, recursive_call;

		recursive = cgraph_edge_recursive_p(edge);
		recursive_call = recursive_call_p(current_function_decl, NODE_DECL(edge->callee));
		gcc_assert(recursive == recursive_call);
		if (recursive)
			return true;
	}

	return false;
}

bool can_convert_to_automatic(gimple_stmt_iterator *gsi,
			      tree size,
			      sbitmap gfp_zones,
			      bool atomic,
			      unsigned HOST_WIDE_INT lhs_align_val)
{
	bool convertible;
	gcall *kalloc_call_stmt;
	tree lhs;
	location_t loc;
	gimple_stmt_ptr use_stmt;
	imm_use_iterator imm_iter;
	sbitmap free_doms;
	basic_block kalloc_bb;

#ifndef TARGET_386
	// TODO: bail out on non-x86 for the time being
	return false;
#endif

	if (!autostack)
		return false;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);

	// supporting these archs will need some more consideration
	if (TREE_INT_CST_LOW(ARCH_KMALLOC_MINALIGN) > PREFERRED_STACK_BOUNDARY / BITS_PER_UNIT)
		return false;

	// bail out on too large objects
	if (TREE_INT_CST_LOW(size) > TREE_INT_CST_LOW(THREAD_SIZE) / 64)
		return false;

	// bail out if not only ZONE_NORMAL is allowed
	if (sbitmap_count_bits(gfp_zones) != 1)
		return false;
	if (!sbitmap_bit_p(gfp_zones, TREE_INT_CST_LOW(ZONE_NORMAL)))
		return false;

	// bail out if DRAP was needed to support this object
	if (lhs_align_val > PREFERRED_STACK_BOUNDARY / BITS_PER_UNIT) {
		if (verbose)
			inform(loc, "autoslab: no stack alignment " HOST_WIDE_INT_PRINT_UNSIGNED, lhs_align_val);
		return false;
	}

	// bail out if GFP_ATOMIC
	if (atomic) {
		if (verbose)
			inform(loc, "autoslab: no stack atomic");
		return false;
	}

	// bail out on self-recursive functions
	if (self_recursive_p()) {
		if (verbose)
			inform(loc, "autoslab: no stack recursive");
		return false;
	}

	// set up free_doms
	free_doms = sbitmap_alloc(last_basic_block_for_fn(cfun));
	sbitmap_clear(free_doms);

	// clean up a little mess that gcc can leave around
	lhs = gimple_call_lhs(kalloc_call_stmt);
	propagate_single_arg_phis(lhs);

	// collect free_doms
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		switch (gimple_code(use_stmt)) {
		default:
			continue;

		case GIMPLE_CALL: {
			tree callee;

			callee = gimple_call_fndecl(use_stmt);
			if (callee == NULL_TREE)
				continue;
			if (callee == kfree_decl || callee == kzfree_decl || callee == kfree_sensitive_decl)
				sbitmap_set_bit(free_doms, gimple_bb(use_stmt)->index);
			continue;
		}

		case GIMPLE_RETURN:
			// bail out if the allocated object is returned
			sbitmap_clear(free_doms);
			break;
		}

		BREAK_FROM_IMM_USE_STMT(imm_iter);
	}

	// bail out if the allocated object is returned or never even freed inside this function
	if (sbitmap_empty_p(free_doms)) {
		convertible = false;
		goto out;
	}

	// expand free_doms with dominating NULL checks
	kalloc_bb = gimple_bb(kalloc_call_stmt);
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		basic_block use_bb;

		if (gimple_code(use_stmt) != GIMPLE_COND)
			continue;

		use_bb = gimple_bb(use_stmt);
		if (use_bb == kalloc_bb)
			continue;
		if (!integer_zerop(gimple_cond_rhs(use_stmt)))
			continue;

		switch (gimple_cond_code(use_stmt)) {
		default:
			debug_gimple_stmt(kalloc_call_stmt);
			debug_gimple_stmt(use_stmt);
			gcc_unreachable();

		case NE_EXPR:
		case EQ_EXPR: {
			sbitmap_iterator sbi;
			unsigned int i;

			gcc_assert(gimple_cond_lhs(use_stmt) == lhs);

			EXECUTE_IF_SET_IN_SBITMAP(free_doms, 0, i, sbi) {
				if (dominated_by_p(CDI_DOMINATORS, BASIC_BLOCK_FOR_FN(cfun, i), use_bb))
					sbitmap_set_bit(free_doms, use_bb->index);
			}
			continue;
		}
		}
	}

	// start the hunt for evidence that the object is not convertible
	convertible = true;

	// for a single free case bail out if it does not post-dominate all non-NULL uses
	if (sbitmap_count_bits(free_doms) == 1) {
		basic_block free_bb, use_bb;

		use_bb = NULL;
		free_bb = BASIC_BLOCK_FOR_FN(cfun, sbitmap_first_set_bit(free_doms));
		FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
			use_bb = gimple_bb(use_stmt);
			if (kalloc_bb == use_bb)
				continue;
			if (sbitmap_bit_p(free_doms, use_bb->index))
				continue;
			if (dominated_by_p(CDI_POST_DOMINATORS, use_bb, free_bb))
				continue;

			convertible = false;
			BREAK_FROM_IMM_USE_STMT(imm_iter);
		}

		goto out;
	}

	// otherwise bail out if a non-NULL use has a path out of the function without encountering a free
	// if only we had multi-vertex dominators...
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		edge e;
		edge_iterator ei;
		basic_block use_bb, *tos, *worklist;

		use_bb = gimple_bb(use_stmt);
		if (kalloc_bb == use_bb)
			continue;
		if (sbitmap_bit_p(free_doms, use_bb->index))
			continue;

		FOR_ALL_BB_FN(use_bb, cfun)
			use_bb->flags &= ~BB_REACHABLE;

		tos = worklist = XNEWVEC(basic_block, n_basic_blocks_for_fn(cfun));

		use_bb = gimple_bb(use_stmt);
		FOR_EACH_EDGE(e, ei, use_bb->succs) {
			if (sbitmap_bit_p(free_doms, e->dest->index))
				continue;

			*tos++ = e->dest;
			e->dest->flags |= BB_REACHABLE;
		}

		while (tos != worklist) {
			basic_block bb;

			bb = *--tos;
			FOR_EACH_EDGE(e, ei, bb->succs) {
				basic_block dest = e->dest;

				if (dest->flags & BB_REACHABLE)
					continue;
				if (sbitmap_bit_p(free_doms, dest->index))
					continue;

				*tos++ = dest;
				dest->flags |= BB_REACHABLE;
			}
		}

		free(worklist);

		if (!(EXIT_BLOCK_PTR_FOR_FN(cfun)->flags & BB_REACHABLE))
			continue;

		convertible = false;
		BREAK_FROM_IMM_USE_STMT(imm_iter);
	}

out:
	sbitmap_free(free_doms);

	if (convertible) {
		if (convert_autostack_decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs to replace %qD", CONVERT_AUTOSTACK, gimple_call_fndecl(kalloc_call_stmt));
			convertible = false;
		}
		if (free_autostack_decl == NULL_TREE) {
			error_at(loc, "autoslab did not find %qs needed for autostack conversion", FREE_AUTOSTACK);
			convertible = false;
		}
	}

	// honour nolocal types and forgo the conversion
	if (convertible && lookup_attribute("nolocal", TYPE_ATTRIBUTES(strip_array_types(TREE_TYPE(lhs)))))
		convertible = false;

	return convertible;
}

/*
 * change the storage duration of an object from allocated to automatic
 */
void autoslab_convert_to_automatic(gimple_stmt_iterator *gsi,
				   tree size,
				   tree flags,
				   sbitmap gfp_zones,
				   unsigned HOST_WIDE_INT lhs_size_val,
				   unsigned HOST_WIDE_INT lhs_align_val)
{
	gcall *kalloc_call_stmt;
	tree block, lhs, type, var, ptr;
	location_t loc;
	gimple_stmt_ptr use_stmt, check_stmt;
	gcall *convert_stmt;
	gphi *phi;
	basic_block cond_bb, join_bb, alloc_bb;
	edge e;
	edge_iterator ei;
	imm_use_iterator imm_iter;
	cgraph_node_ptr callee_node;
	int frequency;

	kalloc_call_stmt = as_a_gcall(gsi_stmt(*gsi));
	loc = gimple_expansion_point_location(kalloc_call_stmt);
	block = gimple_block(kalloc_call_stmt);

	// build the aligned array type for the autostack variable
	gcc_assert(!integer_zerop(size));
	type = build_index_type(fold_build2_loc(loc, MINUS_EXPR, TREE_TYPE(size), size, integer_one_node));
	type = build_array_type(unsigned_char_type_node, type);
	type = build_aligned_type(type, lhs_align_val * BITS_PER_UNIT);
	build_pointer_type(type);

	// build the autostack variable
	var = autoslab_build_local_var(loc, create_tmp_var_name("autostack_var"), type, NULL_TREE, lhs_align_val);
	var = build1_loc(loc, ADDR_EXPR, TYPE_POINTER_TO(type), var);

	// build the autostack ptr
	ptr = autoslab_build_local_var(loc, create_tmp_var_name("autostack_ptr"), TYPE_POINTER_TO(type), NULL_TREE, lhs_align_val);

	// prepare
	convert_stmt = as_a_gcall(gimple_build_call(convert_autostack_decl, 3, var, size, flags));
	gimple_set_location(convert_stmt, loc);
	gimple_set_block(convert_stmt, block);
	gimple_call_set_lhs(convert_stmt, make_ssa_name(ptr, convert_stmt));
	gsi_insert_before(gsi, convert_stmt, GSI_SAME_STMT);
	update_stmt(convert_stmt);

	// update the cgraph
	cond_bb = gimple_bb(convert_stmt);
	callee_node = cgraph_get_create_node(convert_autostack_decl);
	gcc_assert(callee_node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, cond_bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), callee_node, convert_stmt, cond_bb->count, frequency, cond_bb->loop_depth);

	check_stmt = gimple_build_cond(EQ_EXPR, gimple_call_lhs(convert_stmt), null_pointer_node, NULL_TREE, NULL_TREE);
	gimple_set_location(check_stmt, loc);
	gimple_set_block(check_stmt, block);
	gsi_insert_before(gsi, check_stmt, GSI_SAME_STMT);
	update_stmt(check_stmt);

	// split off cond_bb and join_bb
	cond_bb = gimple_bb(check_stmt);
	e = split_block(cond_bb, kalloc_call_stmt);
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALLTHRU;

	e = split_block(cond_bb, check_stmt);
	gcc_assert(cond_bb == e->src);
	alloc_bb = e->dest;
	e->flags = EDGE_TRUE_VALUE;
	e->probability = REG_BR_PROB_VERY_UNLIKELY;
	*gsi = gsi_start_bb(alloc_bb);

	e = make_edge(cond_bb, join_bb, EDGE_FALSE_VALUE);
	e->probability = REG_BR_PROB_VERY_LIKELY;
	join_bb->count = edge_count(e);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	// save then change the return variable of the kalloc stmt
	lhs = gimple_call_lhs(kalloc_call_stmt);
	gimple_call_set_lhs(kalloc_call_stmt, make_ssa_name(ptr, kalloc_call_stmt));
	update_stmt(kalloc_call_stmt);

	// insert phi for lhs
	phi = as_a_gphi(create_phi_node(lhs, join_bb));
	FOR_EACH_EDGE(e, ei, join_bb->preds) {
		if (e->src == alloc_bb)
			add_phi_arg(phi, gimple_call_lhs(kalloc_call_stmt), e, loc);
		else
			add_phi_arg(phi, gimple_call_lhs(convert_stmt), e, loc);
	}
	update_stmt(phi);

	// conditionalize the frees
	FOR_EACH_IMM_USE_STMT(use_stmt, imm_iter, lhs) {
		tree callee;
		gimple_stmt_iterator gsi2, gsi3;
		basic_block erase_bb;
		gcall *erase_stmt;
		gimple_stmt_ptr barrier_stmt;
		edge erase_edge;

		if (gimple_code(use_stmt) != GIMPLE_CALL)
			continue;

		callee = gimple_call_fndecl(use_stmt);
		if (callee == NULL_TREE)
			continue;
		if (callee != kfree_decl && callee != kzfree_decl && callee != kfree_sensitive_decl)
			continue;

		check_stmt = gimple_build_cond(NE_EXPR, gimple_call_lhs(convert_stmt), lhs, NULL_TREE, NULL_TREE);
		gimple_set_location(check_stmt, gimple_location(use_stmt));
		gimple_set_block(check_stmt, gimple_block(use_stmt));
		gsi2 = gsi_for_stmt(use_stmt);
		gsi_insert_before(&gsi2, check_stmt, GSI_SAME_STMT);
		update_stmt(check_stmt);

		// split off cond_bb and join_bb
		cond_bb = gimple_bb(check_stmt);
		e = split_block(cond_bb, use_stmt);
		gcc_assert(cond_bb == e->src);
		join_bb = e->dest;
		e->flags = EDGE_FALLTHRU;

		e = split_block(cond_bb, check_stmt);
		gcc_assert(cond_bb == e->src);
		e->flags = EDGE_TRUE_VALUE;
		e->probability = REG_BR_PROB_VERY_UNLIKELY;

		erase_edge = e = make_edge(cond_bb, join_bb, EDGE_FALSE_VALUE);
		e->probability = REG_BR_PROB_VERY_LIKELY;
		join_bb->count = edge_count(e);

		gcc_assert(dom_info_available_p(CDI_DOMINATORS));
		set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

		// simulate secure erase on the autostack variable
		erase_bb = plugin_split_edge(erase_edge);

		if (callee == kfree_decl)
			erase_stmt = as_a_gcall(gimple_build_call(free_autostack_decl, 2, var, TYPE_SIZE_UNIT(type)));
		else
			erase_stmt = as_a_gcall(gimple_build_call(builtin_decl_implicit(BUILT_IN_MEMSET), 3, var, integer_zero_node, TYPE_SIZE_UNIT(type)));

		gimple_set_location(erase_stmt, gimple_location(use_stmt));
		gimple_set_block(erase_stmt, gimple_block(use_stmt));
		gsi3 = gsi_after_labels(erase_bb);
		gsi_insert_after(&gsi3, erase_stmt, GSI_CONTINUE_LINKING);
		update_stmt(erase_stmt);

		// update the cgraph
		callee_node = cgraph_get_create_node(gimple_call_fndecl(erase_stmt));
		gcc_assert(callee_node);
		frequency = compute_call_stmt_bb_frequency(current_function_decl, erase_bb);
		cgraph_create_edge(cgraph_get_node(current_function_decl), callee_node, erase_stmt, erase_bb->count, frequency, erase_bb->loop_depth);

		if (callee != kfree_decl) {
			gcc_assert(VAR_P(lhs) || (TREE_CODE(lhs) == SSA_NAME && SSA_NAME_VAR(lhs)));
			barrier_stmt = gimple_barrier(lhs, false);
			gimple_set_location(barrier_stmt, gimple_location(use_stmt));
			gimple_set_block(barrier_stmt, gimple_block(use_stmt));
			gsi_insert_after(&gsi3, barrier_stmt, GSI_CONTINUE_LINKING);
		}
	}
}
