/*
 * Copyright 2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "autoslab.h"

regex_t sizeof_regex;

static const char *__get_cache_typename(tree cache)
{
	const char *cache_typename_ptr = NULL;

#if BUILDING_GCC_VERSION >= 4007
	varpool_node_ptr cache_var;
	int idx;
	struct ipa_ref *ref = NULL;

	cache_var = varpool_get_node(cache);
	gcc_assert(cache_var);

	for (idx = 0; ipa_ref_list_referring_iterate(&NODE_SYMBOL(cache_var)->ref_list, idx, ref); idx++) {
		switch (ref->use) {
		default:
			dump_varpool_node(stderr, cache_var);
			gcc_unreachable();

		case IPA_REF_LOAD: {
			imm_use_iterator imm_iter;
			gimple_stmt_ptr cache_load_stmt;
			gcall **candidates;
			size_t i, n;
			tree lhs;

			lhs = gimple_assign_lhs(ref->stmt);
			n = num_imm_uses(lhs);
			candidates = XNEWVEC(gcall *, n);
			i = 0;
			gcc_assert(is_gimple_assign(ref->stmt));

			FOR_EACH_IMM_USE_STMT(cache_load_stmt, imm_iter, lhs) {
				gcall *cache_alloc_stmt;
				tree fndecl;

				if (!is_gimple_call(cache_load_stmt))
					continue;
				cache_alloc_stmt = as_a_gcall(cache_load_stmt);
				fndecl = gimple_call_fndecl(cache_alloc_stmt);
				gcc_assert(fndecl);
				if (!is_cache_alloc(DECL_NAME_POINTER(fndecl)))
					continue;

				gcc_assert(i < n);
				candidates[i++] = cache_alloc_stmt;
			}

			while (i-- && !cache_typename_ptr) {
				tree lhs_type;
				const_tree lhs_typename;

				lhs_type = autoslab_get_alloc_type(gimple_call_lhs(candidates[i]));
				if (!RECORD_OR_UNION_TYPE_P(lhs_type))
					continue;

				lhs_typename = type_name(TYPE_MAIN_VARIANT(lhs_type));
				if (!lhs_typename)
					lhs_typename = type_name(lhs_type);
				gcc_assert(lhs_typename);
				cache_typename_ptr = IDENTIFIER_POINTER(lhs_typename);
			}

			free(candidates);

			if (cache_typename_ptr)
				break;
			else
				continue;
		}

		case IPA_REF_ADDR:
			continue;

		case IPA_REF_STORE:
			continue;
		}
		break;
	}
#endif

	return cache_typename_ptr;
}

static const char *get_cache_typename(gcall *create_stmt)
{
	const char *cache_typename_ptr = NULL;
	tree cache, *candidates;
	gimple_stmt_ptr cache_store_stmt;
	imm_use_iterator imm_iter;
	size_t i, n;

	cache = gimple_call_lhs(create_stmt);
	n = num_imm_uses(cache);
	candidates = XNEWVEC(tree, n);
	i = 0;

	FOR_EACH_IMM_USE_STMT(cache_store_stmt, imm_iter, cache) {
		tree lhs;

		if (!gimple_assign_single_p(cache_store_stmt))
			continue;

		gcc_assert(i < n);
		lhs = gimple_assign_lhs(cache_store_stmt);
		switch (TREE_CODE(lhs)) {
		default:
			debug_gimple_stmt(create_stmt);
			debug_tree(cache);
			debug_gimple_stmt(cache_store_stmt);
			debug_tree(lhs);
			gcc_unreachable();

		case ARRAY_REF:
			continue;

		case COMPONENT_REF:
			continue;

		case MEM_REF:
			continue;

		case SSA_NAME:
			continue;

		case VAR_DECL:
			candidates[i++] = lhs;
			continue;
		}
	}

	while (i-- && !cache_typename_ptr)
		cache_typename_ptr = __get_cache_typename(candidates[i]);

	free(candidates);

	return cache_typename_ptr;
}

// try to extract the typename from sizeof in the src
const char *parse_cache_typename(const char *cache_typename_ptr)
{
	regmatch_t pmatch[3];

	gcc_assert(cache_typename_ptr[0] == 0 && cache_typename_ptr[1] != 0);

	if (regexec(&sizeof_regex, cache_typename_ptr + 1, ARRAY_SIZE(pmatch), pmatch, 0))
		return NULL;

	if (pmatch[2].rm_so == -1)
		return NULL;

	return xstrndup(cache_typename_ptr + 1 + pmatch[2].rm_so, pmatch[2].rm_eo - pmatch[2].rm_so);
}

static void handle_kmem_cache_create_call(gcall *create_stmt)
{
	const char *cache_typename_ptr = NULL;
	tree cache, cache_name, cache_size, cache_typename;
	location_t loc;

	loc = gimple_inline_point_location(create_stmt);

	cache_size = gimple_call_arg(create_stmt, 2);
	cache_typename = gimple_call_arg(create_stmt, 1);
	cache_name = gimple_call_arg(create_stmt, 0);

	switch (TREE_CODE(cache_typename)) {
	default:
		debug_gimple_stmt(create_stmt);
		debug_tree(cache_name);
		debug_tree(cache_typename);
		gcc_unreachable();

	case ADDR_EXPR:
		cache_typename = TREE_OPERAND(cache_typename, 0);
		gcc_assert(TREE_CODE(cache_typename) == STRING_CST);
		cache_typename_ptr = TREE_STRING_POINTER(cache_typename);
		if (!cache_typename_ptr[0])
			break;

		if (verbose)
			inform(loc, "autoslab: typename yes srcstring %qs", TREE_STRING_POINTER(cache_typename));
		return;

	case SSA_NAME:
		if (verbose)
			inform(loc, "autoslab: typename yes srcvar %qE", cache_typename);
		return;
	}
	gcc_assert(cache_typename_ptr);

	cache = gimple_call_lhs(create_stmt);
	if (cache == NULL_TREE)
		return;
	gcc_assert(TREE_CODE(cache) == SSA_NAME);

	switch (TREE_CODE(cache_name)) {
	default:
		debug_gimple_stmt(create_stmt);
		debug_tree(cache_name);
		gcc_unreachable();

	case ADDR_EXPR:
		cache_name = TREE_OPERAND(cache_name, 0);
		switch (TREE_CODE(cache_name)) {
		default:
			debug_gimple_stmt(create_stmt);
			debug_tree(cache_name);
			gcc_unreachable();

		case COMPONENT_REF:
		case STRING_CST:
			break;

		case VAR_DECL:
			if (TREE_READONLY(cache_name)) {
				cache_name = DECL_INITIAL(cache_name);
				gcc_assert(cache_name);
				gcc_assert(TREE_CODE(cache_name) == STRING_CST);
			}
			break;
		}
		break;

	case SSA_NAME:
		break;
	}

	switch (TREE_CODE(cache_size)) {
	default:
		debug_gimple_stmt(create_stmt);
		debug_tree(cache);
		debug_tree(cache_name);
		debug_tree(cache_size);
		gcc_unreachable();

	case INTEGER_CST:
		cache_typename_ptr = parse_cache_typename(cache_typename_ptr);
		if (cache_typename_ptr) {
			if (verbose)
				inform(loc, "autoslab: typename yes parsing %qs", cache_typename_ptr);
			break;
		}
		// FALLTHROUGH

	case SSA_NAME:
		cache_typename_ptr = get_cache_typename(create_stmt);
		if (cache_typename_ptr && verbose)
			inform(loc, "autoslab: typename yes analysis %qs", cache_typename_ptr);
		break;
	}

	if (cache_typename_ptr) {
		cache_typename = build_const_char_string(strlen(cache_typename_ptr) + 1, cache_typename_ptr);
		cache_typename = build1_loc(loc, ADDR_EXPR, TREE_TYPE(gimple_call_arg(create_stmt, 1)), cache_typename);
		gimple_call_set_arg(create_stmt, 1, cache_typename);
		update_stmt(create_stmt);
		return;
	}

	if (verbose) {
		if (TREE_CODE(cache_name) == STRING_CST)
			inform(loc, "autoslab: typename no %qs", TREE_STRING_POINTER(cache_name));
		else
			inform(loc, "autoslab: typename no %qE", cache_name);
	}
}

static unsigned int autotypename_execute(void)
{
	cgraph_node_ptr create_node;
	cgraph_edge_p edge;

	create_node = cgraph_get_node(kmem_cache_create_usercopy_typename_decl);
	gcc_assert(create_node);

	for (edge = create_node->callers; edge; edge = edge->next_caller) {
		tree create_caller;

		create_caller = NODE_DECL(edge->caller);
		gcc_assert(DECL_STRUCT_FUNCTION(create_caller));
		set_current_function_decl(create_caller);

		handle_kmem_cache_create_call(as_a_gcall(edge->call_stmt));

		unset_current_function_decl();
	}
	return 0;
}

static bool autotypename_gate(void)
{
	cgraph_node_ptr node;

	if (!kmem_cache_create_usercopy_typename_decl)
		return false;

	node = cgraph_get_node(kmem_cache_create_usercopy_typename_decl);
	if (!node)
		return false;

	return true;
}

#define PASS_NAME autotypename
#define TODO_FLAGS_START TODO_verify_ssa | TODO_verify_stmts | TODO_update_ssa | TODO_cleanup_cfg | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-simple_ipa-pass.h"

#if (BUILDING_GCC_VERSION >= 9003 && BUILDING_GCC_VERSION < 10000) || BUILDING_GCC_VERSION >= 10001
/*
 * to improve AUTOTYPENAME we stash the size expression in kcalloc/kmalloc_array calls
 * as an extra string argument to kmalloc_typename and convert the underlying STRING_CST's
 * type to the struct/union type, if recognized. this will be consumed by autoslab_convert_call.
 *
 * it has to be done this early as later identifier_global_tag does not work.
 */
static tree __autoslab_resolve_typenames(tree *in, int *walk_subtrees, void *arg)
{
	tree call, fndecl, cache_typename, lhs_type;
	const char *lhs_type_name_ptr;

	call = *in;
	if (TREE_CODE(call) != CALL_EXPR)
		return NULL_TREE;

	fndecl = CALL_EXPR_FN(call);
	if (fndecl == NULL_TREE)
		return NULL_TREE;
	if (TREE_CODE(fndecl) != ADDR_EXPR)
		return NULL_TREE;

	fndecl = TREE_OPERAND(fndecl, 0);
	if (strcmp(DECL_NAME_POINTER(fndecl), "kmalloc_typename"))
		return NULL_TREE;

	// now that we found a kmalloc_typename call, let's try to find the pointee type
	cache_typename = CALL_EXPR_ARG(call, 2);
	gcc_assert(TREE_CODE(cache_typename) == NOP_EXPR);
	cache_typename = TREE_OPERAND(cache_typename, 0);
	gcc_assert(TREE_CODE(cache_typename) == ADDR_EXPR);
	cache_typename = TREE_OPERAND(cache_typename, 0);
	gcc_assert(TREE_CODE(cache_typename) == STRING_CST);
	lhs_type_name_ptr = TREE_STRING_POINTER(cache_typename);
	gcc_assert(!lhs_type_name_ptr[0]);

	lhs_type_name_ptr = parse_cache_typename(lhs_type_name_ptr);
	if (!lhs_type_name_ptr)
		return NULL_TREE;

	// pointee type name found, stash the correponding tree for autoslab_convert_call
	lhs_type = maybe_get_identifier(lhs_type_name_ptr);
	gcc_assert(lhs_type != NULL_TREE);
	lhs_type = identifier_global_tag(lhs_type);

	// triggers on linux-5.18.19/fs/jfs/jfs_dmap.c:dbDiscardAG struct range2trim
	if (lhs_type == NULL_TREE)
		return NULL_TREE;
	build_pointer_type(lhs_type);
	TREE_TYPE(cache_typename) = TYPE_POINTER_TO(lhs_type);

	*walk_subtrees = 0;
	return NULL_TREE;
}

void autoslab_resolve_typenames(void *event_data, void *data __unused)
{
	tree fndecl = (tree)event_data;

	if (fndecl == error_mark_node)
		return;

	gcc_assert(TREE_CODE(fndecl) == FUNCTION_DECL);
	if (DECL_SAVED_TREE(fndecl) == NULL_TREE)
		return;

	walk_tree(&DECL_SAVED_TREE(fndecl), __autoslab_resolve_typenames, NULL, NULL);
}
#endif
