/*
 * Copyright 2011 by Emese Revfy <re.emese@gmail.com>
 * Copyright 2011-2021 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 */

#include "constify.h"

enum states {
	invalid,
	seen_n,
	seen_nm,
	seen_nmo,
	seen_nmoc,
	seen_nmocm,
	seen_no,
	seen_noc,
	seen_nocm,
	num_states
};

enum events {
	not_write,
	mutable_write,
	unsafe_write,
	open_kernel,
	close_kernel,
	num_events
};

struct bb_state {
	gimple_stmt_ptr ins_open;
	gimple_stmt_ptr ins_close;
	gimple_stmt_ptr src_close;
	enum states state;
	bool has_unsafe_write;
	bool changed;
};

GTY(()) tree pax_open_kernel_decl;
bool pax_open_kernel_used_orig;
bool pax_open_kernel_preserved_orig;
GTY(()) tree pax_close_kernel_decl;
bool pax_close_kernel_used_orig;
bool pax_close_kernel_preserved_orig;

static sbitmap pred_states;
static bool tainted[0];

static void dump_bb_state(FILE *dump_file, struct bb_state *bb_state, unsigned int i)
{
	if (bb_state->state == invalid)
		return;

	fprintf(dump_file, "BB %u state: %u unsafe_write %u changed %u\n",
		i, bb_state->state, bb_state->has_unsafe_write, bb_state->changed);
}

static void dump_bb_states(FILE *dump_file, struct bb_state *bb_states)
{
	unsigned int i;

	for (i = 0; i < last_basic_block_for_fn(cfun); i++)
		dump_bb_state(dump_file, &bb_states[i], i);
}

static void gimple_set_not_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, false);
	gimple_set_plf(stmt, GF_PLF_2, false);
}

static bool gimple_set_not_write_p(gimple_stmt_ptr stmt)
{
	return !gimple_plf(stmt, GF_PLF_1) && !gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_mutable_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, false);
	gimple_set_plf(stmt, GF_PLF_2, true);
}

static bool gimple_set_mutable_write_p(gimple_stmt_ptr stmt)
{
	return !gimple_plf(stmt, GF_PLF_1) && gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_unsafe_write(gimple_stmt_ptr stmt)
{
	gimple_set_plf(stmt, GF_PLF_1, true);
	gimple_set_plf(stmt, GF_PLF_2, false);
}

static bool gimple_set_unsafe_write_p(gimple_stmt_ptr stmt)
{
	return gimple_plf(stmt, GF_PLF_1) && !gimple_plf(stmt, GF_PLF_2);
}

static void gimple_set_event(gimple_stmt_ptr stmt, enum events event)
{
	switch (event) {
	case open_kernel:
	case close_kernel:
	case not_write:
		gimple_set_not_write(stmt);
		return;

	case mutable_write:
		gimple_set_mutable_write(stmt);
		return;

	case unsafe_write:
		gimple_set_unsafe_write(stmt);
		return;

	case num_events:
		gcc_unreachable();
	}
}

static bool is_pax_close_kernel(gimple_stmt_ptr stmt);
static bool is_pax_open_kernel(gimple_stmt_ptr stmt);

static enum events gimple_get_event(gimple_stmt_ptr stmt)
{
	if (gimple_set_not_write_p(stmt)) {
		if (is_pax_open_kernel(stmt))
			return open_kernel;
		if (is_pax_close_kernel(stmt))
			return close_kernel;
		return not_write;
	}

	if (gimple_set_mutable_write_p(stmt))
		return mutable_write;

	if (gimple_set_unsafe_write_p(stmt))
		return unsafe_write;

	gcc_unreachable();
}

static bool is_write(gimple_stmt_ptr stmt)
{
	if (gimple_clobber_p(stmt))
		return false;

	return gimple_store_p(stmt);
}

/*
 * determine if type is mutable
 */
enum mutable_type { mutable_yes, mutable_no, mutable_unknown };

static enum mutable_type is_mutable_type(const_tree type)
{
	if (lookup_attribute("no_const", TYPE_ATTRIBUTES(type)))
		return mutable_no;

	if (lookup_attribute("do_const", TYPE_ATTRIBUTES(type)))
		return mutable_yes;

	if (lookup_attribute("mutable_const", TYPE_ATTRIBUTES(type)))
		return mutable_yes;

	return mutable_unknown;
}

/*
 * determine if var is a mutable object
 */
static bool is_mutable_var(const_tree var)
{
	const_tree vartype = TREE_TYPE(var);
	const char *name;

	gcc_assert(VAR_P(var));

	if (!is_global_var(var))
		return false;

	if (is_mutable_type(vartype) == mutable_yes)
		return true;

	if (lookup_attribute("mutable_const", DECL_ATTRIBUTES(var)))
		return true;

	name = get_decl_section_name(var);
	return name && is_const_section(name);
}

/*
 * determine if arg can be a reference to a mutable object
 */
static bool is_mutable_arg(const_tree arg)
{
	switch (TREE_CODE(arg)) {
	default:
		debug_tree(arg);
		gcc_unreachable();

	case ADDR_EXPR:
	case COMPONENT_REF: {
		tree argtype;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
			return true;

		arg = get_inner_arg(TREE_OPERAND(arg, 0));
#if BUILDING_GCC_VERSION == 4005
		if (TREE_CODE(arg) == INDIRECT_REF)
#else
		if (TREE_CODE(arg) == MEM_REF)
#endif
			return is_mutable_type(TREE_TYPE(arg)) == mutable_yes;

		if (TREE_CODE(arg) == VAR_DECL)
			return is_mutable_var(arg);
		return false;
	}

	case INTEGER_CST: {
		tree argtype;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		return is_mutable_type(TREE_TYPE(argtype)) == mutable_yes;
	}

	case SSA_NAME: {
		gimple_stmt_ptr def_stmt;
		tree argtype, rhs1;

		argtype = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(argtype))
			return false;
		if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
			return true;

		def_stmt = SSA_NAME_DEF_STMT(arg);

		if (!is_gimple_assign(def_stmt))
			return false;

		rhs1 = gimple_assign_rhs1(def_stmt);
		switch (gimple_assign_rhs_code(def_stmt)) {
		default:
			fprintf(stderr, "ISMUTABLEARG ASSIGN %s ", get_tree_code_name(gimple_assign_rhs_code(def_stmt)));
			debug_gimple_stmt(def_stmt);
			debug_tree(arg);
			debug_tree(SSA_NAME_VAR(arg));
			debug_tree(rhs1);
			gcc_unreachable();

		case ADDR_EXPR:
		case COMPONENT_REF:
#if 0
			argtype = TREE_TYPE(argtype);
			if (POINTER_TYPE_P(argtype))
				if (is_mutable_type(TREE_TYPE(argtype)) == mutable_yes)
					return true;
#endif

			return is_mutable_arg(rhs1);

		case ARRAY_REF:
		case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#endif
		case PARM_DECL:
			argtype = TREE_TYPE(argtype);
			if (POINTER_TYPE_P(argtype))
				return is_mutable_type(TREE_TYPE(argtype)) == mutable_yes;

			return false;

		case INTEGER_CST:
			return false;

		case CONVERT_EXPR:
		case NOP_EXPR:
			if (!POINTER_TYPE_P(TREE_TYPE(rhs1)))
				return false;
			return is_mutable_arg(rhs1);

		case POINTER_PLUS_EXPR:
		case SSA_NAME:
			return is_mutable_arg(rhs1);

		case VAR_DECL:
			return is_mutable_var(rhs1);
		}
	}

	case VAR_DECL:
		return is_mutable_var(arg);
	}
}

static enum events is_mutable_lhs(gimple_stmt_ptr stmt, tree lhs)
{
	tree ref, offset, reftype;
	poly_int64_pod bitsize, bitpos;
	enum machine_mode mode;
	int unsignedp, reversep, volatilep;

	ref = get_inner_reference(lhs, &bitsize, &bitpos, &offset, &mode, &unsignedp, &reversep, &volatilep, true);
	reftype = TREE_TYPE(ref);

	switch (TREE_CODE(ref)) {
	default:
		print_gimple_stmt(stderr, stmt, 0, TDF_LINENO);
		debug_tree(lhs);
		debug_tree(ref);
		gcc_unreachable();

	case VAR_DECL:
		if (is_mutable_var(ref))
			return mutable_write;
		return INTEGRAL_TYPE_P(TREE_TYPE(lhs)) ? not_write : unsafe_write;

	case INDIRECT_REF:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case TARGET_MEM_REF:
		switch (is_mutable_type(reftype)) {
		default:
			gcc_unreachable();

		case mutable_yes:
			return mutable_write;

		case mutable_no:
			return unsafe_write;

		case mutable_unknown:
			ref = TREE_CODE(ref) == TARGET_MEM_REF ? TMR_BASE(ref) : TREE_OPERAND(ref, 0);
			return is_mutable_arg(ref) ? mutable_write : unsafe_write;
		}

	// simple parameter write
	case PARM_DECL:
		if (!offset)
			return not_write;
		gcc_unreachable();

	// simple return value write
	case RESULT_DECL:
	case SSA_NAME:
		return not_write;
	}
}

enum placement {before, after};

static gimple_stmt_ptr constify_insert_call(tree callee, gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;
	gcall *gcallee;
	cgraph_node_ptr node;
	int frequency;
	basic_block bb;

	// insert call to callee
	stmt = gimple_build_call(callee, 0);
	gimple_set_location(stmt, gimple_location(reference));
	gimple_set_block(stmt, gimple_block(reference));
	gcallee = as_a_gcall(stmt);
	if (placement == before)
		gsi_insert_before(gsi, gcallee, GSI_SAME_STMT);
	else
		gsi_insert_after(gsi, gcallee, GSI_CONTINUE_LINKING);

	// update the cgraph
	bb = gimple_bb(gcallee);
	node = cgraph_get_node(callee);
	gcc_assert(node);
	frequency = compute_call_stmt_bb_frequency(current_function_decl, bb);
	cgraph_create_edge(cgraph_get_node(current_function_decl), node, gcallee, bb->count, frequency, bb->loop_depth);

	return stmt;
}

static bool is_pax_close_kernel(gimple_stmt_ptr stmt)
{
	if (pax_close_kernel_decl == NULL_TREE)
		return false;

	if (!is_gimple_call(stmt))
		return false;

	return gimple_call_fndecl(stmt) == pax_close_kernel_decl;
}

static bool is_pax_open_kernel(gimple_stmt_ptr stmt)
{
	if (pax_open_kernel_decl == NULL_TREE)
		return false;

	if (!is_gimple_call(stmt))
		return false;

	return gimple_call_fndecl(stmt) == pax_open_kernel_decl;
}

static void pax_open_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;

	if (!pax_open_kernel_decl) {
		error_at(gimple_location(reference), "pax_open_kernel must be declared");
		return;
	}

	if (verbose) {
		location_t loc = gsi_end_p(*gsi) ? gimple_location(reference) : gimple_location(gsi_stmt(*gsi));
		inform(loc, "constify: writable access to constified region begins %s",
			placement == before ? "before" : "after");
	}

	stmt = constify_insert_call(pax_open_kernel_decl, gsi, reference, placement);
	gimple_set_event(stmt, open_kernel);
}

static void pax_close_kernel(gimple_stmt_iterator *gsi, gimple_stmt_ptr reference, enum placement placement)
{
	gimple_stmt_ptr stmt;

	if (!pax_close_kernel_decl) {
		error_at(gimple_location(reference), "pax_close_kernel must be declared");
		return;
	}

	if (verbose) {
		location_t loc = gsi_end_p(*gsi) ? gimple_location(reference) : gimple_location(gsi_stmt(*gsi));
		inform(loc, "constify: writable access to constified region ends %s%s",
			placement == before ? "before" : "after", gsi_end_p(*gsi) ? " (approximately)" : "");
	}

	stmt = constify_insert_call(pax_close_kernel_decl, gsi, reference, placement);
	gimple_set_event(stmt, close_kernel);
}

/*
 * determine if stmt can modify a mutable object and thus needs an open kernel to do so
 *
 * a mutable object is either placed into a read-only section or has its type marked with
 * the mutable_const attribute so we have to find pointer dereferences of such types.
 */
static enum events constify_classify_stmt(gimple_stmt_ptr stmt)
{
	tree ref, reftype;

	if (pax_open_kernel_decl && is_pax_open_kernel(stmt))
		return open_kernel;

	if (pax_close_kernel_decl && is_pax_close_kernel(stmt))
		return close_kernel;

	switch (gimple_code(stmt)) {
	default:
		error_at(gimple_location(stmt), "constify: CLASSIFY %s", gimple_code_name[gimple_code(stmt)]);
		debug_gimple_stmt(stmt);
		gcc_unreachable();

	case GIMPLE_ASM: {
		gasm *asm_stmt;
		const char *asm_string;
		size_t i, noutputs;

		asm_stmt = as_a_gasm(stmt);
		asm_string = gimple_asm_string(asm_stmt);
		if (!*asm_string)
			return not_write;

		if (strstr(asm_string, "rap_safe_asm"))
			return not_write;

		return unsafe_write;

		noutputs = gimple_asm_noutputs(asm_stmt);
		for (i = 0; i < noutputs; ++i) {
			tree op, arg;

			op = gimple_asm_output_op(asm_stmt, i);
			arg = TREE_VALUE(op);
			ref = get_inner_arg(arg);
			reftype = TREE_TYPE(ref);

			switch (TREE_CODE(ref)) {
			default:
				debug_gimple_stmt(stmt);
				debug_tree(op);
				debug_tree(arg);
				debug_tree(ref);
				gcc_unreachable();

			case SSA_NAME:
				if (!POINTER_TYPE_P(reftype))
					continue;
				if (is_mutable_type(TREE_TYPE(reftype)) == mutable_yes)
					return mutable_write;
				continue;

			case COMPONENT_REF:
#if BUILDING_GCC_VERSION >= 4006
			case MEM_REF:
#endif
				if (is_mutable_arg(TREE_OPERAND(ref, 0)))
					return mutable_write;
				continue;

			case VAR_DECL:
				if (is_mutable_var(ref))
					return mutable_write;
				continue;
			}
		}

		return unsafe_write;
	}

	case GIMPLE_ASSIGN:
		return is_write(stmt) ? is_mutable_lhs(stmt, gimple_get_lhs(stmt)) : not_write;

	case GIMPLE_CALL: {
		tree lhs, fndecl, count, param;
		const char *fnname;
		unsigned int i, e, flags;

		lhs = gimple_call_lhs(stmt);
		if (lhs && is_mutable_lhs(stmt, lhs) == mutable_write) {
			tree var;
			gimple_stmt_ptr assign;
			gimple_stmt_iterator gsi;

			var = create_tmp_var(TREE_TYPE(lhs), "constify");
			add_referenced_var(var);
			mark_sym_for_renaming(var);

			assign = gimple_build_assign(lhs, var);
			gimple_set_location(assign, gimple_location(stmt));
			gimple_set_block(assign, gimple_block(stmt));
			gsi = gsi_for_stmt(stmt);
			gsi_insert_after(&gsi, assign, GSI_SAME_STMT);

			gimple_call_set_lhs(stmt, var);
			unlink_stmt_vdef(stmt);
			update_stmt(stmt);
			lhs = var;
		}

		fndecl = gimple_call_fndecl(stmt);

		if (fndecl == NULL_TREE)
			return unsafe_write;

		flags = gimple_call_flags(stmt);

		if (flags & ECF_CONST || flags & ECF_PURE)
			return not_write;

#if BUILDING_GCC_VERSION >= 12000
		if (gimple_call_internal_p(stmt, IFN_DEFERRED_INIT))
			return not_write;
#endif

		fnname = DECL_NAME_POINTER(fndecl);
		if (!strcmp(fnname, "memcpy") ||
		    !strcmp(fnname, "memset") ||
		    gimple_call_builtin_p(stmt, BUILT_IN_MEMCPY) ||
		    gimple_call_builtin_p(stmt, BUILT_IN_MEMSET))
			count = gimple_call_arg(stmt, 2);
		else if (!strcmp(fnname, "__memzero"))
			count = gimple_call_arg(stmt, 1);
		else
			return unsafe_write;

		if (integer_zerop(count))
			return not_write;
		else if (TREE_CODE(count) == SSA_NAME) {
			gimple_stmt_ptr def_stmt = SSA_NAME_DEF_STMT(count);
			if (gimple_assign_single_p(def_stmt) && integer_zerop(gimple_assign_rhs1(def_stmt)))
				return not_write;
		}

		return is_mutable_arg(gimple_call_arg(stmt, 0)) ? mutable_write : unsafe_write;

		param = TYPE_ARG_TYPES(gimple_call_fntype(stmt));
		if (param == NULL_TREE)
			return unsafe_write;

		for (i = 0, e = gimple_call_num_args(stmt); i < e; ++i, param = param ? TREE_CHAIN(param) : NULL_TREE) {
			tree arg;

			arg = gimple_call_arg(stmt, i);
			gcc_assert(!param || POINTER_TYPE_P(TREE_TYPE(arg)) == POINTER_TYPE_P(TREE_VALUE(param)));
			if (!POINTER_TYPE_P(TREE_TYPE(arg)))
				continue;
			if (param && is_mutable_type(TREE_TYPE(TREE_VALUE(param))) == mutable_yes)
				return unsafe_write;
			if (is_mutable_type(TREE_TYPE(TREE_TYPE(arg))) == mutable_yes)
				return unsafe_write;
			if (CONSTANT_CLASS_P(arg))
				continue;
			if (is_mutable_arg(arg))
				return unsafe_write;
			continue;
		}

		return not_write;
	}

	case GIMPLE_COND:
	case GIMPLE_DEBUG:
	case GIMPLE_GOTO:
	case GIMPLE_LABEL:
	case GIMPLE_NOP:
	case GIMPLE_PREDICT:
	case GIMPLE_RETURN:
	case GIMPLE_SWITCH:
		return not_write;
	}

	gcc_unreachable();
}

static gimple_stmt_ptr constify_remove_stmt(gimple_stmt_ptr stmt)
{
	gimple_stmt_iterator gsi;

	location_t loc;
	tree block;

	gsi = gsi_for_stmt(stmt);
	loc = gimple_location(stmt);
	block = gimple_block(stmt);

	unlink_stmt_vdef(stmt);

	stmt = gimple_build_nop();
	gimple_set_location(stmt, loc);
	gimple_set_block(stmt, block);
	gsi_replace(&gsi, stmt, true);
	gimple_set_event(stmt, not_write);

	return stmt;
}

/*
 * the workhorse: find and instrument mutable writes that would fail otherwise at runtime
 *
 * the instrumentation surrounds the mutable writes with pax_open_kernel/pax_close_kernel
 * calls. there's some light effort at optimizing the number and placement of these calls
 * as their performance impact is unfortunately not negligible with current processor features.
 */
static enum events constify_step(struct bb_state *bb_state, gimple_stmt_ptr stmt, enum events event)
{
	switch (event) {
	case not_write:
		return not_write;

	case mutable_write:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			bb_state->state = seen_nm;

			bb_state->ins_open = stmt;
			bb_state->ins_close = stmt;
			return mutable_write;

		case seen_no:
		case seen_nmo:
			return mutable_write;

		case seen_nm:
		case seen_nocm:
		case seen_nmocm:
			bb_state->ins_close = stmt;
			return mutable_write;

		case seen_noc:
			bb_state->state = seen_nocm;

			gcc_assert(bb_state->src_close);
			gcc_assert(stmt != bb_state->src_close);
			gcc_assert(bb_state->ins_close == NULL);

			constify_remove_stmt(bb_state->src_close);
			bb_state->ins_close = stmt;
			bb_state->src_close = NULL;

			bb_state->changed = true;
			return mutable_write;

		case seen_nmoc:
			bb_state->state = seen_nmocm;

			gcc_assert(bb_state->src_close);
			gcc_assert(stmt != bb_state->src_close);

			constify_remove_stmt(bb_state->src_close);
			bb_state->ins_close = stmt;
			bb_state->src_close = NULL;

			bb_state->changed = true;
			return mutable_write;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case unsafe_write:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			return unsafe_write;

		case seen_nm: {
			gimple_stmt_iterator gsi2;

			bb_state->state = seen_n;

			gcc_assert(bb_state->ins_open);
			gsi2 = gsi_for_stmt(bb_state->ins_open);
			pax_open_kernel(&gsi2, stmt, before);
			bb_state->ins_open = NULL;

			gcc_assert(bb_state->ins_close);
			gsi2 = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi2, stmt, after);
			bb_state->ins_close = NULL;

			bb_state->changed = true;
			return unsafe_write;
		}

		case seen_no:
		case seen_nmo:
			return mutable_write;

		case seen_noc:
		case seen_nmoc:
			bb_state->state = seen_n;

			gcc_assert(bb_state->src_close);
			bb_state->src_close = NULL;
			gcc_assert(bb_state->ins_close == NULL);
			return unsafe_write;

		case seen_nocm:
		case seen_nmocm: {
			gimple_stmt_iterator gsi2;

			bb_state->state = seen_n;

			gcc_assert(bb_state->ins_close);
			gsi2 = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi2, stmt, after);
			bb_state->ins_close = NULL;

			bb_state->changed = true;
			return unsafe_write;
		}

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case open_kernel:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
			bb_state->state = seen_no;
			return open_kernel;

		case seen_nm:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			bb_state->ins_close = NULL;

			bb_state->changed = true;
			return open_kernel;

		case seen_no:
		case seen_nmo:
			error_at(gimple_location(stmt), "constify: pax_open_kernel encountered after a previous pax_open_kernel");
			gcc_unreachable();

		case seen_noc:
			bb_state->state = seen_no;

			constify_remove_stmt(stmt);

			gcc_assert(bb_state->src_close);
			constify_remove_stmt(bb_state->src_close);
			bb_state->src_close = NULL;

			gcc_assert(bb_state->ins_close == NULL);

			bb_state->changed = true;
			return open_kernel;

		case seen_nocm:
			bb_state->state = seen_no;

			constify_remove_stmt(stmt);
			gcc_assert(bb_state->ins_close);
			bb_state->ins_close = NULL;

			bb_state->changed = true;
			return open_kernel;

		case seen_nmoc:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			gcc_assert(!bb_state->ins_close);

			bb_state->changed = true;
			return open_kernel;

		case seen_nmocm:
			bb_state->state = seen_nmo;

			constify_remove_stmt(stmt);
			gcc_assert(bb_state->ins_close);
			bb_state->ins_close = NULL;

			bb_state->changed = true;
			return open_kernel;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case close_kernel:
		switch (bb_state->state) {
		case invalid:
			gcc_unreachable();

		case seen_n:
		case seen_nm:
			error_at(gimple_location(stmt), "constify: pax_close_kernel encountered without a previous pax_open_kernel");
			gcc_unreachable();

		case seen_noc:
		case seen_nmoc:
		case seen_nocm:
		case seen_nmocm:
			error_at(gimple_location(stmt), "constify: pax_close_kernel encountered after a previous pax_close_kernel");
			gcc_unreachable();

		case seen_no:
			bb_state->state = seen_noc;

			gcc_assert(!bb_state->src_close);
			bb_state->src_close = stmt;
			return close_kernel;

		case seen_nmo:
			bb_state->state = seen_nmoc;

			gcc_assert(!bb_state->src_close);
			bb_state->src_close = stmt;
			gcc_assert(!bb_state->ins_close);
			return close_kernel;

		case num_states:
			gcc_unreachable();
		}
		gcc_unreachable();

	case num_events:
		gcc_unreachable();
	}
	gcc_unreachable();
}

/*
 * the main analysis run over all BBs in reverse postorder to determine where to place
 * potential open/close instrumentation or emit some of them directly when there's no
 * choice for placement.
 */
static struct bb_state *analyze_mutable_writes(int *postorder, unsigned int postorder_num)
{
	unsigned int i;
	struct bb_state *bb_states;

	i = last_basic_block_for_fn(cfun);
	bb_states = XNEWVEC(struct bb_state, i);
	for (; i; i--) {
		bb_states[i - 1].ins_open = NULL;
		bb_states[i - 1].ins_close = NULL;
		bb_states[i - 1].src_close = NULL;
		bb_states[i - 1].state = invalid;
		bb_states[i - 1].has_unsafe_write = false;
		bb_states[i - 1].changed = false;
	}

	for (i = postorder_num; i; i--) {
		unsigned int j;
		basic_block bb;
		edge e;
		edge_iterator ei;
		gimple_stmt_ptr stmt;
		gimple_stmt_iterator gsi;
		struct bb_state *bb_state;

		j = postorder[i - 1];
		bb = BASIC_BLOCK_FOR_FN(cfun, j);
		bb_state = &bb_states[j];

		sbitmap_clear(pred_states);

		FOR_EACH_EDGE(e, ei, bb->preds) {
			enum states state;

			state = bb_states[e->src->index].state;
			if (state == invalid)
				continue;

			sbitmap_set_bit(pred_states, state);
		}

		switch (sbitmap_count_bits(pred_states)) {
		case 0:
			gcc_assert(j == ENTRY_BLOCK || j == EXIT_BLOCK);
			bb_state->state = seen_n;
			break;

		case 1: {
			enum states state = (enum states)sbitmap_first_set_bit(pred_states);

			switch (state) {
			default:
				debug_bb(bb);
				debug_sbitmap(pred_states);
				gcc_unreachable();

			case seen_n:
			case seen_no:
			case seen_nmo:
				bb_state->state = state;
				break;

			case seen_nmocm:
				gcc_assert(j == EXIT_BLOCK);
				bb_state->state = state;
				break;

			case seen_nm:
			case seen_nocm:
				FOR_EACH_EDGE(e, ei, bb->preds) {
					struct bb_state *bb_state_pred = &bb_states[e->src->index];

					if (bb_state_pred->state == invalid)
						continue;

					gcc_assert(bb_state_pred->ins_close);
					constify_step(bb_state_pred, bb_state_pred->ins_close, unsafe_write);
				}
				bb_state->state = seen_n;
				break;

			case seen_noc:
			case seen_nmoc:
				FOR_EACH_EDGE(e, ei, bb->preds) {
					struct bb_state *bb_state_pred = &bb_states[e->src->index];

					if (bb_state_pred->state == invalid)
						continue;

					gcc_assert(!bb_state_pred->ins_close);
					constify_step(bb_state_pred, NULL, unsafe_write);
				}
				bb_state->state = seen_n;
				break;
			}
			break;
		}

		default:
			FOR_EACH_EDGE(e, ei, bb->preds) {
				enum states state;
				struct bb_state *bb_state_pred = &bb_states[e->src->index];

				state = bb_state_pred->state;
				switch (state) {
				default:
					debug_sbitmap(pred_states);
					gcc_unreachable();

				case seen_n:
					continue;

				case seen_nm:
					gcc_assert(bb_state_pred->ins_close);
					constify_step(bb_state_pred, bb_state_pred->ins_close, unsafe_write);
					continue;

				case seen_noc:
				case seen_nmoc:
					gcc_assert(!bb_state_pred->ins_close);
					constify_step(bb_state_pred, NULL, unsafe_write);
					continue;
				}
			}

			bb_state->state = seen_n;
			break;
		}

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			enum events event;

			stmt = gsi_stmt(gsi);
			event = constify_classify_stmt(stmt);
			gimple_set_event(stmt, constify_step(bb_state, stmt, event));

			if (event != unsafe_write)
				continue;

			bb_state->has_unsafe_write = true;

			if (!bb->loop_father)
				continue;

			bb->loop_father->aux = tainted;
		}

		if (verbose && dump_file)
			dump_bb_state(dump_file, bb_state, j);
	}

	return bb_states;
}

static basic_block loop_single_exit_block(loop_p loop)
{
	struct loop_exit *exit;
	basic_block bb_single_exit;

	exit = loop->exits->next;
	if (!exit->e)
		return NULL;

	bb_single_exit = exit->e->dest;
	for (exit = exit->next; exit->e; exit = exit->next)
		if (bb_single_exit != exit->e->dest)
			return NULL;

	return bb_single_exit;
}

static bool instrument_mutable_writes(int *postorder, unsigned int postorder_num, struct bb_state *bb_states)
{
	bool changed = false;
	unsigned int i;
#if BUILDING_GCC_VERSION < 12000
	loop_p loop;
#endif
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

	for (i = postorder_num; i; i--) {
		unsigned int j;
		gimple_stmt_iterator gsi;
		struct bb_state *bb_state;

		j = postorder[i - 1];
		bb_state = &bb_states[j];

		if (bb_state->changed)
			changed = true;

		if (bb_state->ins_open) {
			gsi = gsi_for_stmt(bb_state->ins_open);
			pax_open_kernel(&gsi, bb_state->ins_open, before);
			bb_state->ins_open = NULL;
			changed = true;
		}

		if (bb_state->ins_close) {
			gsi = gsi_for_stmt(bb_state->ins_close);
			pax_close_kernel(&gsi, bb_state->ins_close, after);
			bb_state->ins_close = NULL;
			changed = true;
		}
	}

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_FROM_INNERMOST) {
#else
	FOR_EACH_LOOP(loop, LI_FROM_INNERMOST) {
#endif
		basic_block *bbs;
		unsigned int opens, closes;
		gimple_stmt_iterator gsi;
		basic_block bb_open, bb_close;
		gimple_stmt_ptr stmt;

		if (loop->aux == tainted)
			continue;

		opens = closes = 0;
		bbs = get_loop_body(loop);

		for (i = 0; i < loop->num_nodes; i++) {
			for (gsi = gsi_start_bb(bbs[i]); !gsi_end_p(gsi); gsi_next(&gsi)) {
				stmt = gsi_stmt(gsi);
				switch (gimple_get_event(stmt)) {
				default:
					continue;

				case unsafe_write:
					gcc_unreachable();

				case open_kernel:
					constify_remove_stmt(stmt);
					opens++;
					continue;

				case close_kernel:
					constify_remove_stmt(stmt);
					closes++;
					continue;
				}
			}
		}

		free(bbs);

		gcc_assert(opens == closes);
		if (opens == 0 && closes == 0)
			continue;

		bb_open = get_immediate_dominator(CDI_DOMINATORS, loop->header);
		gcc_assert(single_succ_p(bb_open));
		gsi = gsi_last_nondebug_bb(bb_open);
		if (gsi_end_p(gsi)) {
			gcc_assert(single_succ_p(bb_open));
			stmt = gsi_stmt(gsi_last_nondebug_bb(single_succ(bb_open)));
		} else
			stmt = gsi_stmt(gsi);
		gcc_assert(stmt);
		pax_open_kernel(&gsi, stmt, gsi_end_p(gsi) ? before : after);

		bb_close = loop_single_exit_block(loop);
		if (bb_close && dominated_by_p(CDI_DOMINATORS, bb_close, bb_open)) {
			gcc_assert(dominated_by_p(CDI_POST_DOMINATORS, bb_open, bb_close));
			gsi = gsi_start_nondebug_after_labels_bb(bb_close);
			gcc_assert(!gsi_end_p(gsi));
			pax_close_kernel(&gsi, gsi_stmt(gsi), before);
		} else {
			unsigned int j;
			edge e;
#if BUILDING_GCC_VERSION <= 4007
			VEC(edge, heap) *exits = get_loop_exit_edges(loop);

			FOR_EACH_VEC_ELT(edge, exits, j, e) {
#else
			VEC(edge, heap) exits = get_loop_exit_edges(loop);

			FOR_EACH_VEC_ELT(exits, j, e) {
#endif
				basic_block new_close;

				gcc_assert(e->src->loop_father == loop);

				gsi = gsi_last_nondebug_bb(e->dest);
				gcc_assert(!gsi_end_p(gsi));
				stmt = gsi_stmt(gsi);

				new_close = plugin_split_edge(e);
				gcc_assert(single_pred_p(new_close));
				gsi = gsi_start_nondebug_after_labels_bb(new_close);
				pax_close_kernel(&gsi, stmt, after);
			}
#if BUILDING_GCC_VERSION <= 4007
			VEC_free(edge, heap, exits);
#elif BUILDING_GCC_VERSION < 11000
			exits.release();
#endif
		}

		changed = true;
	}

	return changed;
}

static void propagate_loop_taint(int *postorder, unsigned int postorder_num)
{
#if BUILDING_GCC_VERSION < 12000
	loop_p loop;
#endif
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_FROM_INNERMOST) {
#else
	FOR_EACH_LOOP(loop, LI_FROM_INNERMOST) {
#endif
		loop_p outer;

		if (loop->aux != tainted)
			continue;

		for (outer = loop_outer(loop); outer; outer = loop_outer(outer))
			outer->aux = tainted;
	}
}

static unsigned int instrument_mutable_writes_execute(void)
{
	unsigned int round;
	bool changed;
#if BUILDING_GCC_VERSION < 12000
	loop_p loop;
#endif
#if BUILDING_GCC_VERSION <= 4008
	loop_iterator li;
#endif

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_INCLUDE_ROOT)
#else
	FOR_EACH_LOOP(loop, LI_INCLUDE_ROOT)
#endif
		loop->aux = NULL;

	pred_states = sbitmap_alloc(num_states);

	/*
	 * the core loop that iterates over analysis and instrumentation.
	 */
	changed = true;
	for (round = 0; changed; round++) {
		int *postorder;
		unsigned int postorder_num;
		struct bb_state *bb_states;

		if (verbose && dump_file)
			fprintf(dump_file, "\n;; start round %u for %s\n", round, DECL_NAME_POINTER(current_function_decl));

		postorder = XNEWVEC(int, n_basic_blocks_for_fn(cfun));
		postorder_num = post_order_compute(postorder, true, false);
		gcc_assert(ENTRY_BLOCK == postorder[postorder_num - 1]);

		bb_states = analyze_mutable_writes(postorder, postorder_num);

		if (round == 0)
			propagate_loop_taint(postorder, postorder_num);

		changed = instrument_mutable_writes(postorder, postorder_num, bb_states);

		if (verbose && dump_file) {
			dump_bb_states(dump_file, bb_states);

			dump_flow_info(dump_file, dump_flags);
			fprintf(dump_file, "\n;; end round %u for %s (%s)\n\n",
				round, DECL_NAME_POINTER(current_function_decl), changed ? "changed" : "unchanged");
		}

		XDELETEVEC(bb_states);
		XDELETEVEC(postorder);
	}

	if (verbose && round > 1 && DECL_DECLARED_INLINE_P(current_function_decl))
		inform(expansion_point_location(DECL_SOURCE_LOCATION(current_function_decl)), "constify: instrumented possibly inlined function %qD", current_function_decl);

	sbitmap_free(pred_states);

#if BUILDING_GCC_VERSION <= 4008
	FOR_EACH_LOOP(li, loop, LI_INCLUDE_ROOT)
#else
	FOR_EACH_LOOP(loop, LI_INCLUDE_ROOT)
#endif
		loop->aux = NULL;

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static bool instrument_mutable_writes_gate(void)
{
	if (MODULE != NULL_TREE)
		return true;

	if (has_trait(current_function_decl, "boot"))
		return false;

	return true;
}

#define PASS_NAME instrument_mutable_writes
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_remove_unused_locals | TODO_update_ssa | TODO_cleanup_cfg | TODO_ggc_collect | TODO_rebuild_cgraph_edges | TODO_verify_flow
#include "gcc-generate-gimple-pass.h"
