/*
 * Copyright 2012-2019 by PaX Team <pageexec@freemail.hu>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"

static void rap_hash_function(const_tree fntype, rap_hash_flags_t flags, unsigned char sip_hash[8]);

static const unsigned char rap_hash_tree_code[MAX_TREE_CODES] = {
	[0] = 0,
	[1] = 0,
	[2] = 0,
	[3] = 0,
	[4] = 0,
	[OFFSET_TYPE] = 10,
	[ENUMERAL_TYPE] = 20,
	[BOOLEAN_TYPE] = 30,
	[INTEGER_TYPE] = 40,
	[REAL_TYPE] = 50,
	[POINTER_TYPE] = 60,
	[REFERENCE_TYPE] = 70,
#if BUILDING_GCC_VERSION >= 4006
	[NULLPTR_TYPE] = 80,
#endif
	[FIXED_POINT_TYPE] = 0,
	[COMPLEX_TYPE] = 100,
	[VECTOR_TYPE] = 110,
	[ARRAY_TYPE] = 120,
	[RECORD_TYPE] = 130,
	[UNION_TYPE] = 140,
	[QUAL_UNION_TYPE] = 0,
	[VOID_TYPE] = 160,
#if BUILDING_GCC_VERSION >= 5000 && BUILDING_GCC_VERSION < 9000
	[POINTER_BOUNDS_TYPE] = 170,
#endif
	[FUNCTION_TYPE] = 180,
	[METHOD_TYPE] = 0,
	[LANG_TYPE] = 0,
};

static void rap_fold_hash(unsigned char *out, const unsigned char *in, unsigned long long inlen)
{
	static const unsigned char rap_sip_key[16] = {
		'P', 'a', 'X', ' ', 'T', 'e', 'a', 'm',
		'R', 'A', 'P', ' ', 'H', 'A', 'S', 'H',
	};

	siphash24fold(out, in, inlen, rap_sip_key);
}

// compute the final hash value in the range [1,INT_MAX]
// the % and +1 trick leaves the value 0 available for marking non-indirectly callable functions
// and INT_MIN (0x80000000) for longjmp targets (sign extended)
// return places will use the (sign extended) range [INT_MIN+1,-1] ([0x8000001,0xffffffff])
static rap_hash_t rap_extract_hash(const unsigned char sip_hash[8])
{
	rap_hash_t hash;
	unsigned long long dividend, divisor;

	memcpy(&dividend, sip_hash, sizeof dividend);
//	divisor = 1ULL << (sizeof hash * 8 - 1);
//	divisor |= divisor - 1;
	divisor = 0x7fffffffUL;
	hash.hash = dividend % divisor + 1;
	return hash;
}

static const_tree main_variant_name(const_tree type)
{
	const_tree name;

	name = type_name(TYPE_MAIN_VARIANT(type));

	// handle typedefs of anonymous enums/structs/unions
	if (name == NULL_TREE) {
		tree t;

		// assume that the first variant of an otherwise anonymous type has the name we want
		// the first variant is the last in the chain of variants
		for (t = TYPE_MAIN_VARIANT(type); t; t = TYPE_NEXT_VARIANT(t)) {
			name = type_name(t);
#if BUILDING_GCC_VERSION >= 9000
			if (lookup_attribute("first_variant", TYPE_ATTRIBUTES(t))) {
				gcc_assert(name);
				break;
			}
#endif
		}
	}

	return name;
}

static void rap_hash_tree(const_tree type, rap_hash_flags_t flags, unsigned char sip_hash[8]);

static void rap_hash_type_name(const_tree type, rap_hash_flags_t flags, unsigned char sip_hash[8])
{
	const_tree name;

	name = main_variant_name(type);

	// in some situations the anonymous type isn't aliased directly at all (e.g., only a pointer to it is defined)
	// in such cases assume that its definition must be present in all TUs and thus its fields can be used in the type hash
	if (name) {
		gcc_assert(TREE_CODE(name) == IDENTIFIER_NODE);
		rap_fold_hash(sip_hash, (const unsigned char *)IDENTIFIER_POINTER(name), IDENTIFIER_LENGTH(name));
	} else {
		const_tree field;

		switch (TREE_CODE(type)) {
		default:
			debug_tree(type);
			gcc_unreachable();

		case ENUMERAL_TYPE:
			gcc_assert(TYPE_VALUES(type));
			for (field = TYPE_VALUES(type); field; field = TREE_CHAIN(field)) {
				const_tree purpose, value;
				HOST_WIDE_INT enumval;

				purpose = TREE_PURPOSE(field);
				gcc_assert(TREE_CODE(purpose) == IDENTIFIER_NODE);
				rap_fold_hash(sip_hash, (const unsigned char *)IDENTIFIER_POINTER(purpose), IDENTIFIER_LENGTH(purpose));

				value = TREE_VALUE(field);
				gcc_assert(TREE_CODE(value) == INTEGER_CST);
				enumval = tree_to_shwi(value);
				rap_fold_hash(sip_hash, (const unsigned char *)&enumval, sizeof enumval);
			}

			break;

		case RECORD_TYPE:
		case UNION_TYPE:
			gcc_assert(TYPE_FIELDS(type));
			for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
				gcc_assert(TREE_CODE(field) == FIELD_DECL);
				rap_hash_tree(TREE_TYPE(field), flags, sip_hash);
			}

			break;
		}
	}
}

static void rap_hash_type_precision(const_tree type, unsigned char sip_hash[8])
{
	unsigned HOST_WIDE_INT size;

	gcc_assert(TYPE_PRECISION(type));

	size = TYPE_PRECISION(type);
	rap_fold_hash(sip_hash, (const unsigned char *)&size, sizeof size);
}

// the core computation of the rap hash
// the first piece is a (hopefully) compiler independent encoding of the type, derived from the gcc tree code
// the second piece is type specific information, such as the size, qualifiers, (recursively) referenced types, etc
static void rap_hash_tree(const_tree type, rap_hash_flags_t flags, unsigned char sip_hash[8])
{
	enum tree_code code;
	unsigned int attrs;

	code = TREE_CODE(type);
	attrs = rap_hash_tree_code[code];
	if (!attrs) {
		fprintf(stderr, "unhandled tree_code %s %d\n", get_tree_code_name(code), code);
		debug_tree(type);
		gcc_unreachable();
	}
	rap_fold_hash(sip_hash, (const unsigned char *)&attrs, sizeof attrs);

	enum {
	// attrs layout for
		// - all types:
		RAP_HASH_VOLATILE		= 1U << 31,
		RAP_HASH_NOT_VOLATILE		= 1U << 30,
		RAP_HASH_CONST			= 1U << 29,
		RAP_HASH_NOT_CONST		= 1U << 28,

		// - pointer types:
		RAP_HASH_RESTRICT		= 1U << 27,
		RAP_HASH_NOT_RESTRICT		= 1U << 26,

		// - C integer types:
		RAP_HASH_UNSIGNED		= 1U << 25,
		RAP_HASH_SIGNED			= 1U << 24,

		RAP_HASH_UNQUALIFIED_CHAR	= 1U << 23,
		RAP_HASH_CHAR			= 1U << 22,
		RAP_HASH_SHORT			= 1U << 21,
		RAP_HASH_INT			= 1U << 20,
		RAP_HASH_LONG			= 1U << 19,
		RAP_HASH_LONG_LONG		= 1U << 18,
		RAP_HASH_INT128			= 1U << 17,
		RAP_HASH_WCHAR			= 1U << 16,
		RAP_HASH_CHAR16			= 1U << 15,
		RAP_HASH_CHAR32			= 1U << 14,

		// - C float types
		RAP_HASH_FLOAT			= 1U << 13,
		RAP_HASH_DOUBLE			= 1U << 12,
		RAP_HASH_LONG_DOUBLE		= 1U << 11,
		RAP_HASH_DFLOAT32		= 1U << 10,
		RAP_HASH_DFLOAT64		= 1U << 9,
		RAP_HASH_DFLOAT128		= 1U << 8,
#if BUILDING_GCC_VERSION >= 7000
		RAP_HASH_FLOAT16		= 1U << 7,
		RAP_HASH_FLOAT32		= 1U << 6,
		RAP_HASH_FLOAT64		= 1U << 5,
		RAP_HASH_FLOAT128		= 1U << 4,
		RAP_HASH_FLOAT32X		= 1U << 3,
		RAP_HASH_FLOAT64X		= 1U << 2,
		RAP_HASH_FLOAT128X		= 1U << 1,
#endif
	};

	attrs = 0;
	if (flags.qual_volatile)
		attrs |= TYPE_VOLATILE(type) ? RAP_HASH_VOLATILE : RAP_HASH_NOT_VOLATILE;
	if (flags.qual_const)
		attrs |= TYPE_READONLY(type) ? RAP_HASH_CONST : RAP_HASH_NOT_CONST;

	switch (code) {
	default:
		debug_tree(type);
		gcc_unreachable();
		break;

	case VOID_TYPE:
		break;

	case OFFSET_TYPE:
		rap_hash_tree(TREE_TYPE(type), flags, sip_hash);
		rap_hash_tree(TYPE_OFFSET_BASETYPE(type), flags, sip_hash);
		break;

	case FUNCTION_TYPE:
		rap_hash_function(type, flags, sip_hash);
		break;

	case RECORD_TYPE:
		rap_hash_type_name(type, flags, sip_hash);
		break;

	case UNION_TYPE:
		rap_hash_type_name(type, flags, sip_hash);
		break;

#if BUILDING_GCC_VERSION >= 5000 && BUILDING_GCC_VERSION < 9000
	case POINTER_BOUNDS_TYPE:
		rap_hash_type_name(type, flags, sip_hash);
		rap_hash_type_precision(type, sip_hash);
		break;
#endif

	case POINTER_TYPE:
	case REFERENCE_TYPE:
		rap_hash_tree(TREE_TYPE(type), flags, sip_hash);
		break;

	case VECTOR_TYPE:
		rap_hash_tree(TREE_TYPE(type), flags, sip_hash);
		rap_hash_type_precision(TREE_TYPE(type), sip_hash);
		break;

	case COMPLEX_TYPE:
		rap_hash_tree(TREE_TYPE(type), flags, sip_hash);
		break;

	case ARRAY_TYPE:
		rap_hash_tree(TREE_TYPE(type), flags, sip_hash);
		break;

	case REAL_TYPE: {
		const_tree main_variant = TYPE_MAIN_VARIANT(type);

		switch (TYPE_PRECISION(main_variant)) {
		default:
			debug_tree(type);
			debug_tree(TYPE_MAIN_VARIANT(type));
			gcc_unreachable();

		case 16:
//			attrs |= RAP_HASH_FLOAT;
			break;

		case 32:
//			attrs |= RAP_HASH_FLOAT;
			break;

		case 64:
//			attrs |= RAP_HASH_DOUBLE;
			break;

		case 80:
		case 128:
			attrs |= RAP_HASH_LONG_DOUBLE;
			break;
		}
		rap_hash_type_precision(main_variant, sip_hash);
		break;
	}

	case ENUMERAL_TYPE:
		rap_hash_type_name(type, flags, sip_hash);
		// FALLTHROUGH
	case BOOLEAN_TYPE:
		rap_hash_type_precision(type, sip_hash);
		break;

	case INTEGER_TYPE: {
		attrs |= TYPE_UNSIGNED(type) ? RAP_HASH_UNSIGNED : RAP_HASH_SIGNED;
		rap_hash_type_precision(type, sip_hash);
		break;
	}
	}

	rap_fold_hash(sip_hash, (const unsigned char *)&attrs, sizeof attrs);
}

static const_tree rap_dequal_argtype(const_tree argtype)
{
	// since gcc/tree.c:free_lang_data_in_type removes const/volatile from the top level param decl
	// we have to simulate it here as this can be called earlier from the frontend as well
	if (TYPE_READONLY(argtype) || TYPE_VOLATILE(argtype)) {
		int quals;

		quals = TYPE_QUALS(argtype) & ~TYPE_QUAL_CONST & ~TYPE_QUAL_VOLATILE;
		argtype = build_qualified_type(CONST_CAST_TREE(argtype), quals);
	}

	return argtype;
}

// main function to compute the rap hash for function types
// while virtual class methods are always replaced with their ancestor,
// callers can decide whether to fully utilize that information via flags.method_ancestor
static void rap_hash_function(const_tree fntype, rap_hash_flags_t flags, unsigned char sip_hash[8])
{
	function_args_iterator args_iter;
	const_tree arg;

	switch (TREE_CODE(fntype)) {
	default:
		debug_tree(fntype);
		gcc_unreachable();

	case FUNCTION_TYPE:
		// 1. hash the result
		rap_hash_tree(TREE_TYPE(fntype), flags, sip_hash);

		// 2. hash the function parameters
		FOREACH_FUNCTION_ARGS(fntype, arg, args_iter) {
			const_tree argtype = arg;

			argtype = rap_dequal_argtype(argtype);
			rap_hash_tree(argtype, flags, sip_hash);
		}
		break;
	}
}

static rap_hash_t rap_hash_function_type(const_tree fntype, rap_hash_flags_t flags)
{
	unsigned char sip_hash[8] = { };
	rap_hash_t hash;

	rap_hash_function(fntype, flags, sip_hash);
	hash = rap_extract_hash(sip_hash);

	gcc_assert(hash.hash);
	return hash;
}

static rap_hash_t rap_hash_function_decl(const_tree fndecl, rap_hash_flags_t flags)
{
	tree fntype;

	gcc_assert(TREE_CODE(fndecl) == FUNCTION_DECL);
	fntype = TREE_TYPE(fndecl);

	switch (TREE_CODE(fntype)) {
	default:
		debug_tree(fndecl);
		gcc_unreachable();

	case FUNCTION_TYPE: {
		unsigned char sip_hash[8] = { };
		rap_hash_t hash;

		rap_hash_function(fntype, flags, sip_hash);
		if (flags.func_name)
			rap_fold_hash(sip_hash, (const unsigned char *)DECL_NAME_POINTER(fndecl), DECL_NAME_LENGTH(fndecl));
//		if (flags.func_scope)
//			rap_hash_tree(current_scope, flags, sip_hash);
		hash = rap_extract_hash(sip_hash);

		gcc_assert(hash.hash);
		return hash;
	}
	}
}

rap_hash_t rap_lookup_imprecise_rap_hash(const_tree fn)
{
	const_tree attr = NULL_TREE;
	rap_hash_t hash;

	if (DECL_P(fn)) {
		attr = lookup_attribute("rap_hash_override", DECL_ATTRIBUTES(fn));
		fn = TREE_TYPE(fn);
	}
	if (!attr)
		attr = lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fn));

	if (!attr) {
		// handle special late generated dtors/ctors from ASAN/etc that don't get a type hash
		// so far the type seems to be void(void) only, so safe to compute the hash this late
		gcc_assert(TREE_CODE(fn) == FUNCTION_TYPE);
		gcc_assert(TREE_TYPE(fn) == void_type_node);
		gcc_assert(TREE_CODE(TYPE_ARG_TYPES(fn)) == TREE_LIST);
		gcc_assert(TREE_VALUE(TYPE_ARG_TYPES(fn)) == TREE_VALUE(void_list_node));
		gcc_assert(!TREE_CHAIN(TYPE_ARG_TYPES(fn)));
		return rap_hash_function_type(fn, imprecise_rap_hash_flags);
	}

	attr = TREE_VALUE(TREE_VALUE(attr));
	gcc_assert(TREE_CODE(attr) == INTEGER_CST);
	hash.hash = TREE_INT_CST_LOW(attr);
	return hash;
}

/*
rap_hash_t rap_lookup_precise_rap_hash(const_tree fndecl)
{
	const_tree attr;
	rap_hash_t hash;

	gcc_assert(DECL_P(fndecl));
	attr = lookup_attribute("rap_hash", DECL_ATTRIBUTES(fndecl));
	gcc_assert(attr);

	attr = TREE_VALUE(TREE_VALUE(attr));
	gcc_assert(TREE_CODE(attr) == INTEGER_CST);
	hash.hash = TREE_INT_CST_LOW(attr);
	return hash;
}
*/

static void rap_calculate_fntype_hash(tree fntype, tree vindex)
{
	tree attr, args;
	rap_hash_t hash;

	gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);

	attr = lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fntype));
	if (attr) {
		fntype = TYPE_CANONICAL(fntype);
		if (!fntype)
			return;
		if (lookup_attribute("rap_hash", TYPE_ATTRIBUTES(TYPE_CANONICAL(fntype))))
			return;
		add_type_attr(fntype, "rap_hash", TREE_VALUE(attr));
		return;
	}

	hash = rap_hash_function_type(fntype, imprecise_rap_hash_flags);
	args = tree_cons(NULL_TREE, build_int_cst_type(rap_hash_type_node, hash.hash), NULL_TREE);
	add_type_attr(fntype, "rap_hash", args);
}

void rap_calculate_fndecl_hash(tree fndecl)
{
	tree attr, args, fntype;
	rap_hash_t decl_hash, type_hash;

	fntype = TREE_TYPE(fndecl);
	gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);

	decl_hash = rap_hash_function_decl(fndecl, precise_rap_hash_flags);
	attr = lookup_attribute("rap_hash", DECL_ATTRIBUTES(fndecl));
	args = tree_cons(NULL_TREE, build_int_cst_type(rap_hash_type_node, decl_hash.hash), NULL_TREE);
	if (attr) {
		rap_hash_t old_hash;

		old_hash.hash = TREE_INT_CST_LOW(TREE_VALUE(TREE_VALUE(attr)));
		if (decl_hash.hash != old_hash.hash) {
			DECL_ATTRIBUTES(fndecl) = copy_list(DECL_ATTRIBUTES(fndecl));
			attr = lookup_attribute("rap_hash", DECL_ATTRIBUTES(fndecl));
			gcc_assert(attr);
			TREE_VALUE(attr) = args;
		}
	} else {
		DECL_ATTRIBUTES(fndecl) = copy_list(DECL_ATTRIBUTES(fndecl));
		DECL_ATTRIBUTES(fndecl) = tree_cons(get_identifier("rap_hash"), args, DECL_ATTRIBUTES(fndecl));
	}

	type_hash = rap_hash_function_decl(fndecl, imprecise_rap_hash_flags);
	attr = lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fntype));
	if (attr) {
		rap_hash_t old_hash;

		old_hash.hash = TREE_INT_CST_LOW(TREE_VALUE(TREE_VALUE(attr)));
		if (type_hash.hash != old_hash.hash) {
			if (!strchr(DECL_NAME_POINTER(fndecl), '.'))
				inform(DECL_SOURCE_LOCATION(fndecl), "fixed RAP hash mismatch on %qD (%x vs. %x)", fndecl, old_hash.hash, type_hash.hash);
			args = tree_cons(NULL_TREE, build_int_cst_type(rap_hash_type_node, type_hash.hash), NULL_TREE);
			update_type_attr(fntype, "rap_hash", args);
			attr = lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fntype));
		}

		fntype = TYPE_CANONICAL(fntype);
		if (!fntype)
			return;
		if (lookup_attribute("rap_hash", TYPE_ATTRIBUTES(fntype)))
			return;
		add_type_attr(fntype, "rap_hash", TREE_VALUE(attr));
		return;
	}

#if BUILDING_GCC_VERSION >= 7000 && BUILDING_GCC_VERSION < 10000
	// workaround for https://gcc.gnu.org/bugzilla/show_bug.cgi?id=92817
	if (DECL_SOURCE_LOCATION(fndecl) == BUILTINS_LOCATION && !strcmp(DECL_NAME_POINTER(fndecl), "__gcov_exit")) {
		tree fntype = TREE_TYPE(fndecl);

		gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);
		gcc_assert(TREE_TYPE(fntype) == void_type_node);
		gcc_assert(TREE_CODE(TYPE_ARG_TYPES(fntype)) == TREE_LIST);
		gcc_assert(TREE_VALUE(TYPE_ARG_TYPES(fntype)) == TREE_VALUE(void_list_node));
		gcc_assert(TREE_CODE(TREE_CHAIN(TYPE_ARG_TYPES(fntype))) == TREE_LIST);
		gcc_assert(TREE_VALUE(TREE_CHAIN(TYPE_ARG_TYPES(fntype))) == TREE_VALUE(void_list_node));

		TREE_TYPE(fndecl) = build_function_type_list(void_type_node, NULL);
	}
#endif

	args = tree_cons(NULL_TREE, build_int_cst_type(rap_hash_type_node, type_hash.hash), NULL_TREE);
	add_type_attr(fntype, "rap_hash", args);
}

static void rap_calculate_param_hashes(tree fntype)
{
	tree param;

	for (param = TYPE_ARG_TYPES(fntype); param; param = TREE_CHAIN(param)) {
		fntype = TREE_VALUE(param);
		if (TREE_CODE(fntype) != POINTER_TYPE)
			continue;

		fntype = TREE_TYPE(fntype);
		if (TREE_CODE(fntype) != FUNCTION_TYPE)
			continue;

		rap_calculate_fntype_hash(fntype, NULL_TREE);
	}
}

void rap_calculate_func_hashes(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;

	FOR_EACH_FUNCTION(node) {
		const_tree fndecl;
		tree fntype;

		if (NODE_GLOBAL(node)->inlined_to)
			continue;

		fndecl = NODE_DECL(node);
		gcc_assert(fndecl);

		rap_calculate_fndecl_hash(CONST_CAST_TREE(fndecl));

		fntype = TREE_TYPE(fndecl);
		gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);
		rap_calculate_param_hashes(fntype);
	}
}

static tree rap_hash_walk_tree(tree *tp, int *walk_subtrees, void *data)
{
	tree fntype = *tp;
	struct walk_stmt_info *wi = (struct walk_stmt_info *)data;
#if BUILDING_GCC_VERSION >= 7000
	gimple_stmt_ptr stmt = wi ? wi->stmt : NULL;
#else
	gimple_stmt_ptr stmt = wi ? gsi_stmt(wi->gsi) : NULL;
#endif

	switch (TREE_CODE(fntype)) {
	default:
		break;
		if (stmt)
			debug_gimple_stmt(stmt);
		debug_tree(fntype);
		gcc_unreachable();

	case INTEGER_TYPE:
		break;

	case FUNCTION_DECL:
		if (lookup_attribute("rap_hash", TYPE_ATTRIBUTES(TREE_TYPE(fntype))))
			break;
		if (stmt)
			debug_gimple_stmt(stmt);
		debug_tree(fntype);
		gcc_unreachable();

	case ADDR_EXPR:
	case ARRAY_TYPE:
	case FIELD_DECL:
#if BUILDING_GCC_VERSION >= 4006
	case MEM_REF:
#endif
	case PARM_DECL:
	case POINTER_TYPE:
	case SSA_NAME:
	case TARGET_MEM_REF:
	case VAR_DECL:
		fntype = TREE_TYPE(fntype);
		if (TREE_CODE(fntype) == POINTER_TYPE)
			fntype = TREE_TYPE(fntype);
		if (TREE_CODE(fntype) != FUNCTION_TYPE)
			break;
		// FALLTHROUGH

	case FUNCTION_TYPE:
		rap_calculate_fntype_hash(fntype, NULL_TREE);
		break;
	}

	return NULL_TREE;
}

void rap_calculate_fptr_hashes(void *event_data __unused, void *data __unused)
{
	cgraph_node_ptr node;

	gcc_assert(!in_lto_p);

	// find all indirect calls and store the fptr hash for them
	FOR_EACH_FUNCTION_WITH_GIMPLE_BODY(node) {
		basic_block bb;

		gcc_assert(DECL_STRUCT_FUNCTION(NODE_DECL(node)));
		push_cfun(DECL_STRUCT_FUNCTION(NODE_DECL(node)));
#if BUILDING_GCC_VERSION <= 4007
		current_function_decl = NODE_DECL(node);
#endif

		FOR_EACH_BB_FN(bb, cfun) {
			gimple_stmt_iterator gsi;

			for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
				tree fptr, fntype, fndecl;
				gimple_stmt_ptr stmt;

				stmt = gsi_stmt(gsi);
				if (!is_gimple_call(stmt))
					continue;

				if (is_inexpensive_builtin(gimple_call_fndecl(stmt)))
					continue;

				fptr = gimple_call_fn(stmt);
				if (!fptr || TREE_CODE(fptr) == INTEGER_CST)
					continue;

				fntype = TREE_TYPE(fptr);
				gcc_assert(POINTER_TYPE_P(fntype));

				fntype = TREE_TYPE(fntype);
				gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);
				rap_calculate_param_hashes(fntype);

				if (fntype != gimple_call_fntype(stmt)) {
					fntype = gimple_call_fntype(stmt);
					gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);
					rap_calculate_param_hashes(fntype);
				}

				switch (TREE_CODE(fptr)) {
				default:
					debug_gimple_stmt(stmt);
					debug_tree(fptr);
					debug_tree(TREE_TYPE(fptr));
					debug_tree(fntype);
					gcc_unreachable();

				case ADDR_EXPR:
					fndecl = TREE_OPERAND(fptr, 0);

					if (fntype != TREE_TYPE(fndecl))
						rap_calculate_fntype_hash(TREE_TYPE(fndecl), NULL_TREE);
					rap_calculate_fntype_hash(fntype, NULL_TREE);
					break;

				case PARM_DECL:
				case SSA_NAME:
				case VAR_DECL:
					fntype = TREE_TYPE(fptr);
					gcc_assert(TREE_CODE(fntype) == POINTER_TYPE);
					fntype = TREE_TYPE(fntype);
					gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);
					rap_calculate_fntype_hash(fntype, NULL_TREE);
					if (fntype != gimple_call_fntype(stmt)) {
						rap_calculate_fntype_hash(gimple_call_fntype(stmt), NULL_TREE);
						gcc_assert(rap_lookup_imprecise_rap_hash(fntype).hash == rap_lookup_imprecise_rap_hash(gimple_call_fntype(stmt)).hash);
					}
					break;
				}

			}
		}

		pop_cfun();
#if BUILDING_GCC_VERSION <= 4007
		current_function_decl = NULL_TREE;
#endif
	}

	FOR_EACH_FUNCTION_WITH_GIMPLE_BODY(node) {
		basic_block bb;

		gcc_assert(DECL_STRUCT_FUNCTION(NODE_DECL(node)));
		push_cfun(DECL_STRUCT_FUNCTION(NODE_DECL(node)));
#if BUILDING_GCC_VERSION <= 4007
		current_function_decl = NODE_DECL(node);
#endif

		FOR_EACH_BB_FN(bb, cfun) {
			gimple_stmt_iterator gsi;

			for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
				gimple_stmt_ptr stmt;
				struct walk_stmt_info wi = {};

				stmt = gsi_stmt(gsi);
				switch (gimple_code(stmt)) {
				default:
					continue;

				case GIMPLE_ASSIGN:
				case GIMPLE_CALL:
					walk_gimple_stmt(&gsi, NULL, rap_hash_walk_tree, &wi);
					continue;
				}
			}
		}

		pop_cfun();
#if BUILDING_GCC_VERSION <= 4007
		current_function_decl = NULL_TREE;
#endif
	}
}

static void __rap_hash_finish_type(tree type)
{
	tree field;

	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field))
		if (RECORD_OR_UNION_TYPE_P(TREE_TYPE(field)))
			__rap_hash_finish_type(TREE_TYPE(field));
		else
			walk_tree_without_duplicates(&TREE_TYPE(field), rap_hash_walk_tree, NULL);
}

void rap_hash_finish_type(void *event_data, void *data __unused)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	switch (TREE_CODE(type)) {
	default:
		debug_tree(type);
		gcc_unreachable();

#if BUILDING_GCC_VERSION >= 5000
	case ENUMERAL_TYPE:
		return;
#endif

	case RECORD_TYPE:
	case UNION_TYPE:
		if (TYPE_FIELDS(type) == NULL_TREE)
			return;

		__rap_hash_finish_type(type);
		return;
	}
}

void rap_hash_compute_attr(void *event_data, void *data __unused)
{
	tree fndecl = (tree)event_data;

	if (fndecl == error_mark_node)
		return;

	if (TREE_CODE(fndecl) != FUNCTION_DECL)
		return;

	if (TREE_CODE(TREE_TYPE(fndecl)) != FUNCTION_TYPE)
		return;

	rap_calculate_fndecl_hash(fndecl);
}
