/*
 * Copyright 2022 by PaX Team <pageexec@freemail.hu>, Pawel Wieczorkiewicz <wipawel@grsecurity.net>
 * Licensed under the GPL v2
 *
 * Homepage: http://pax.grsecurity.net/
 */

#include "rap.h"
#include "gcc-ksyms.h"

static void emit_call_sls_barrier(rtx_insn *insn)
{
#ifdef GCC_DF_H
#ifdef TARGET_386
	rtx xorbody, parallel;
#if BUILDING_GCC_VERSION >= 5000
	df_ref use;
#else
	df_ref *use;
#endif
	const_tree fn;
	bitmap live_in;
	basic_block bb;
	rtx_insn *cur_insn;
	bool live;

	// Ignore calls where AX is live (e.g. indirect call *%rax)
	FOR_EACH_INSN_USE(use, insn) {
#if BUILDING_GCC_VERSION >= 5000
		if (DF_REF_REGNO(use) == AX_REG)
#else
		if (DF_REF_REGNO(*use) == AX_REG)
#endif
			return;
	}

	// ignore calls returning void after which AX cannot be live
	fn = get_call_target(insn);
	if (fn) {
		tree fntype = strip_array_types(TREE_TYPE(fn));

		switch (TREE_CODE(fn)) {
		default:
			debug_rtx(insn);
			debug_tree(fn);
			debug_tree(TREE_TYPE(fn));
			gcc_unreachable();

#if BUILDING_GCC_VERSION >= 4006
		case MEM_REF:
#endif
		case TARGET_MEM_REF:
			if (POINTER_TYPE_P(fntype))
				fntype = TREE_TYPE(fntype);
			break;

		case ARRAY_REF:
		case COMPONENT_REF:
		case FIELD_DECL:
		case PARM_DECL:
		case SSA_NAME:
		case VAR_DECL:
			gcc_assert(FUNCTION_POINTER_TYPE_P(fntype));
			fntype = TREE_TYPE(fntype);
			break;

		case FUNCTION_DECL:
			break;
		}
		gcc_assert(TREE_CODE(fntype) == FUNCTION_TYPE);

		if (TREE_TYPE(fntype) == void_type_node)
			return;
	}

	// ignore calls where AX is already set by the SSP prologue sequence
	bb = BLOCK_FOR_INSN(insn);
	live_in = BITMAP_ALLOC(NULL);
	bitmap_clear(live_in);
	bitmap_copy(live_in, df_get_live_in(bb));
	df_simulate_initialize_forwards(bb, live_in);

	// - simulate insns forward in the BB until we reach the call insn
	FOR_BB_INSNS(bb, cur_insn) {
		int value;

		if (cur_insn == insn)
			break;

		if (!INSN_P(cur_insn))
			continue;

		if (CALL_P(cur_insn)) {
			bitmap_clear_bit(live_in, AX_REG);
			continue;
		}

		// - first see if the first part of the simulation finds AX live
		df_simulate_find_noclobber_defs(cur_insn, live_in);
		live = bitmap_bit_p(live_in, AX_REG);

		// - now run the full simulation
		df_simulate_one_insn_forwards(bb, cur_insn, live_in);

		// - if the full simulation removed liveness, see if it was because of the SSP prologue
		if (live == bitmap_bit_p(live_in, AX_REG))
			continue;
		gcc_assert(live && !bitmap_bit_p(live_in, AX_REG));

		// - the write to AX in the SSP prologue does not make it live for gcc but it is for our purposes
		if (!find_regno_note(cur_insn, REG_UNUSED, AX_REG))
			continue;

		if (find_regno_note(cur_insn, REG_DEAD, AX_REG))
			continue;

		value = INSN_CODE(cur_insn);
		if (value < 0)
			continue;
		if (strncmp(get_insn_name(value), "stack_tls_protect_set", sizeof("stack_tls_protect_set") - 1) &&
		    strncmp(get_insn_name(value), "stack_protect_set", sizeof("stack_protect_set") - 1))
			continue;

		bitmap_set_bit(live_in, AX_REG);
	}

	// omit the xor if we find AX live at the call
	live = bitmap_bit_p(live_in, AX_REG);
	BITMAP_FREE(live_in);
	if (live)
		return;

	// emit "xor %eax, %eax"
	xorbody = gen_rtx_set(gen_rtx_REG(SImode, AX_REG), gen_rtx_XOR(SImode, gen_rtx_REG(SImode, AX_REG), gen_rtx_REG(SImode, AX_REG)));
	parallel = gen_rtx_PARALLEL(SImode, rtvec_alloc(2));
	XVECEXP(parallel, 0, 0) = xorbody;
	XVECEXP(parallel, 0, 1) = gen_rtx_CLOBBER(VOIDmode, gen_rtx_REG(CCmode, FLAGS_REG));
	emit_insn_before_setloc(parallel, insn, INSN_LOCATION(insn));
#else
#error unsupported target
#endif
#endif
}

rtx_insn *emit_jump_sls_barrier(rtx_insn *insn, unsigned int loc)
{
	rtx_insn *stuffing;

#ifdef TARGET_386
	rtx int3;

	int3 = gen_rtx_ASM_OPERANDS(VOIDmode, ggc_strdup("int3"), empty_string, 0, rtvec_alloc(0), rtvec_alloc(0), rtvec_alloc(0), loc);
	MEM_VOLATILE_P(int3) = 1;
	stuffing = emit_insn_after(int3, insn);

	return emit_barrier_after(stuffing);
#else
	start_sequence();
	expand_builtin_trap();
	stuffing = get_insns();
	end_sequence();
	return emit_insn_after(stuffing, insn);
#endif
}

static rtx_insn *rap_handle_sls_call(rtx_insn *insn)
{
	rtx body;

	body = PATTERN(insn);
	if (GET_CODE(body) == PARALLEL)
		body = XVECEXP(body, 0, 0);
	if (GET_CODE(body) == SET)
		body = SET_SRC(body);
	gcc_assert(GET_CODE(body) == CALL);

	body = XEXP(body, 0);
	gcc_assert(MEM_P(body));

	body = XEXP(body, 0);
	switch (GET_CODE(body)) {
	default:
		print_rtl_single(stderr, insn);
		print_rtl_single(stderr, body);
		gcc_unreachable();

	case MEM:
	case REG:
		break;

	case SYMBOL_REF: {
		tree fn;
		unsigned int loc;

		gcc_assert(SYMBOL_REF_FUNCTION_P(body));
		fn = SYMBOL_REF_DECL(body);
		if (fn == NULL_TREE)
			break;

		gcc_assert(TREE_CODE(fn) == FUNCTION_DECL);

		if (!SIBLING_CALL_P(insn))
			break;

		gcc_assert(SYMBOL_REF_FUNCTION_P(body));
		loc = INSN_LOCATION(insn);

#if BUILDING_GCC_VERSION >= 4007 && BUILDING_GCC_VERSION < 8000
		for (insn = NEXT_INSN(insn); NEXT_INSN(insn); insn = NEXT_INSN(insn)) {
			if (GET_CODE(insn) == BARRIER)
				continue;
			if (GET_CODE(insn) == NOTE)
				continue;
			insn = PREV_INSN(insn);
			break;
		}
		gcc_assert(insn);
#endif

		return emit_jump_sls_barrier(insn, loc);
	}
	}

	emit_call_sls_barrier(insn);
	if (SIBLING_CALL_P(insn) && !rap_retpoline_gate()) {
		// put a trap after indirect jumps to stop speculation past them
		return emit_jump_sls_barrier(insn, INSN_LOCATION(insn));
	}

	return insn;
}

// no computed_goto_p() for RTX; simplified version, not for general use!
static bool is_computed_goto(rtx_insn *insn)
{
	rtx set = pc_set(insn);

	if (!set)
		return false;

	switch (GET_CODE(SET_SRC(set))) {
	default:
		print_rtl_single(stderr, insn);
		gcc_unreachable();

	case IF_THEN_ELSE:
	case LABEL_REF:
		return false;

	case MEM:
		break;
	}

	return true;
}

// stop speculation after ...
static bool branch_needs_sls_p(rtx_insn *insn)
{
	// ... unconditional jumps,
	if (any_uncondjump_p(insn))
		return true;

	if (rap_retpoline_gate())
		return false;

	// ... returns,
	if (returnjump_p(insn))
		return true;

	// ... switch table jumps,
	if (tablejump_p(insn, NULL, NULL))
		return true;

	// ... computed gotos
	if (is_computed_goto(insn))
		return true;

	return false;
}

static unsigned int rap_sls_execute(void)
{
	rtx_insn *insn;

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);
#ifdef GCC_DF_H
	df_finish_pass(true);
	df_scan_add_problem();
	df_scan_alloc(NULL);
	df_scan_blocks();
	df_note_add_problem();
	df_analyze();
#endif

	for (insn = get_insns(); insn; insn = NEXT_INSN(insn)) {
		if (INSN_DELETED_P(insn))
			continue;

		switch (GET_CODE(insn)) {
		case JUMP_INSN:
			if (branch_needs_sls_p(insn))
				insn = emit_jump_sls_barrier(insn, INSN_LOCATION(insn));
			break;

		case CALL_INSN:
			// rtl match (call_insn (set (reg) (call (mem))))
			insn = rap_handle_sls_call(insn);
			break;

		default:
			break;
		}
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	return 0;
}

static bool rap_sls_gate(void)
{
	if (MODULE != NULL_TREE)
		return true;

	if (has_trait(current_function_decl, "boot"))
		return false;

	return true;
}

#define PASS_NAME rap_sls
#define TODO_FLAGS_FINISH TODO_dump_func | TODO_verify_rtl_sharing | TODO_df_verify | TODO_df_finish
#include "gcc-generate-rtl-pass.h"
