/*
 * Copyright 2022 by Open Source Security, Inc.
 *
 * Licensed under the GPL v2
 *
 * Implement the 'canary' struct member attribute.
 *
 * This attribute can be put on structure members with an integer type. It
 * requires an argument containing the canary value to check objects against.
 *
 * Code making use of such typed objects will get the first read to such an
 * object instrumented and the annotated members checked to have one of the
 * allowed values. If not, a call to __builtin_trap() happens.
 *
 * We ignore writes to such objects on purpose, to not require special wrappers
 * for initializing the canary on object allocation.
 *
 * TODO:
 * - add canary member automatically (annotate the type only)
 * - use automatically generated symbolic names instead of generating them
 *   during build via scripts
 * - support multiple canary values, not only one, to support transient
 *   initializing/dying states
 * - IPA pass to handle ADDR_EXPR call arguments
 */

#include "gcc-common.h"

static struct plugin_info type_canary_plugin_info = {
	.version	= "20221003",
	.help		= "verbose\t\tverbose mode\n"
};

__visible int plugin_is_GPL_compatible;

static int verbose;

static tree handle_canary_attr(tree *node, tree name, tree args, int flags, bool *no_add_attrs)
{
	tree type, decl = *node;
	location_t loc;

	*no_add_attrs = true;

	gcc_assert(DECL_P(decl));
	loc = DECL_SOURCE_LOCATION(decl);

	if (TREE_CODE(decl) != FIELD_DECL) {
		error_at(loc, "attribute %qE applies to structure members only", name);
		return NULL_TREE;
	}

	type = TREE_TYPE(decl);
	if (TREE_CODE(type) != INTEGER_TYPE) {
		error_at(loc, "attribute %qE applies to integer types only", name);
		return NULL_TREE;
	}

	args = TREE_VALUE(args);
	if (args == NULL_TREE || TREE_CODE(args) != INTEGER_CST) {
		error_at(loc, "attribute %qE requires an integer argument", name);
		return NULL_TREE;
	}

	/* TODO: Check the full list, once we support more than one value. */
	if (!int_fits_type_p(args, type)) {
		error_at(loc, "%qE value %qE does not fit in %qT", name, args, type);
		return NULL_TREE;
	}

	/*
	 * Further checks like verifying the upper type is a RECORD_TYPE (instead
	 * of a UNION_TYPE) need to be deferred to finish_type(), when the type is
	 * complete.
	 */
	*no_add_attrs = false;

	return NULL_TREE;
}

static struct attribute_spec canary_attr_spec = ATTRIBUTE_SPEC_INIT("canary", 1, 1 /* TODO: support a list, not only one */, true, false, false, handle_canary_attr, true, NULL);

static void register_attributes(void *event_data, void *data)
{
	register_attribute(&canary_attr_spec);
}

static tree get_canary_field(const_tree type)
{
	tree attr;
	gcc_assert(TYPE_P(type));
	attr = lookup_attribute("*has_canary", TYPE_ATTRIBUTES(type));
	return attr ? TREE_VALUE(attr) : NULL_TREE;
}

static void set_canary_field(tree type, tree decl)
{
	gcc_assert(TYPE_P(type) && TREE_CODE(decl) == FIELD_DECL);
	gcc_assert(!lookup_attribute("*has_canary", TYPE_ATTRIBUTES(type)));
	add_type_attr(type, "*has_canary", decl);
}

static bool is_canary_type(const_tree type)
{
	gcc_assert(TYPE_P(type));
	return get_canary_field(type) != NULL_TREE;
}

static void type_check_canary_attr(tree type)
{
	tree canary_field = NULL_TREE;
	tree field;

	/*
	 * We have to allow unions here too, to verify none of its members has the
	 * 'canary' attribute.
	 */
	if (!RECORD_OR_UNION_TYPE_P(type))
		return;

	/*
	 * Most of the sanity checks have already been done by handle_canary_attr().
	 * However, it only saw the annotated member itself, out of context. Do the
	 * remaining context checks here:
	 * - verify a given type has only one canary member at max and
	 * - a canary member isn't part of a union (even if it has only a single
	 *   member, to keep things simple).
	 */
	for (field = TYPE_FIELDS(type); field; field = TREE_CHAIN(field)) {
		tree field_type = TREE_TYPE(field);
		tree inner_canary = NULL_TREE;

		if (RECORD_OR_UNION_TYPE_P(field_type)) {
			inner_canary = get_canary_field(field_type);
			if (!inner_canary)
				continue;

		} else {
			if (!lookup_attribute("canary", DECL_ATTRIBUTES(field)))
				continue;

			gcc_assert(!AGGREGATE_TYPE_P(field_type));
			inner_canary = field;
		}

		gcc_assert(TREE_CODE(inner_canary) == FIELD_DECL);

		/*
		 * Sanity of the 'canary field' was already checked (being an integer
		 * type), now verify it's not part of a union.
		 */
		if (TREE_CODE(type) != RECORD_TYPE) {
			location_t loc = DECL_SOURCE_LOCATION(field);
			bool nested = inner_canary != field;
			const_tree name = type_name(type);

			if (name)
				error_at(loc, "%s%<canary%> %s cannot be part of %<union %E%>", nested ? "nested " : "", nested ? "type" : "member", name);
			else
				error_at(loc, "%s%<canary%> %s cannot be part of anonymous union", nested ? "nested " : "", nested ? "type" : "member");

			if (nested) {
				tree tmp;

				tmp = inner_canary;
				while ( (tmp = get_canary_field(TREE_TYPE(tmp))) )
					inner_canary = tmp;

				inform(DECL_SOURCE_LOCATION(inner_canary), "nested %<canary%> member is:");
			}

			continue;
		}

		if (canary_field) {
			error_at(DECL_SOURCE_LOCATION(field), "a type can have only one %<canary%> member");
			inform(DECL_SOURCE_LOCATION(canary_field), "previous use was here:");
			continue;
		}

		canary_field = field;
	}

	if (canary_field)
		set_canary_field(type, canary_field);
}

/*
 * We could have used an unused C type flag like TYPE_LANG_FLAG_6 (_4 and _5
 * already taken by CONSTIFY and STRUCTLEAK) but that doesn't really scale with
 * the rate we keep on writing new plugins. ;)  So we use a bitmap instead!
 */
static bitmap typemap;

static void finish_type(void *event_data, void *data)
{
	tree type = (tree)event_data;

	if (type == NULL_TREE || type == error_mark_node)
		return;

	gcc_assert(TYPE_P(type));

	/* Ignore forward declarations. */
	if (!COMPLETE_TYPE_P(type))
		return;

	/* Allocate type bitmap on demand. */
	if (!typemap) {
		bitmap_obstack_initialize(NULL);
		typemap = BITMAP_ALLOC(NULL);
	}

	/* We see types multiple times, just walk them once. */
	if (bitmap_bit_p(typemap, TYPE_UID(type)))
		return;

	type_check_canary_attr(type);

	/* Skip this type next time we see it. */
	bitmap_set_bit(typemap, TYPE_UID(type));
}

static void typemap_cleanup(void *event_data, void *data)
{
	if (!typemap)
		return;

	BITMAP_FREE(typemap);
	bitmap_obstack_release(NULL);
	typemap = NULL;
}

static tree canary_from_ref(tree mem_ref, location_t loc)
{
	tree cref, field, type;

	gcc_assert(TREE_CODE(mem_ref) == MEM_REF);

	cref = unshare_expr(mem_ref);
	type = TREE_TYPE(mem_ref);

	/*
	 * find_uninstrumented_ptr() might run into narrowed MEM_REFs that use a
	 * narrow type.
	 * Detect these and upcast the unshared MEM_REF to the canary type for
	 * building the canary member chain.
	 */
	if (!is_canary_type(type)) {
		tree op, ptype;

		op = TREE_OPERAND(cref, 0);
		gcc_assert(TREE_CODE(op) == SSA_NAME);

		ptype = TREE_TYPE(op);
		gcc_assert(POINTER_TYPE_P(ptype));

		// change the mem_ref to have the canary type
		type = TREE_TYPE(ptype);
		TREE_TYPE(cref) = type;
		TYPE_READONLY(type) = 1;

		// change the offset to zero for building up the field_decl chain below
		TREE_OPERAND(cref, 1) = build_int_cst(ptype, 0);
	}

//	fprintf(stderr, ">>> mref = "); debug_tree(mem_ref);
//	fprintf(stderr, ">>> cref = "); debug_tree(cref);

	gcc_assert(is_canary_type(type));

	/*
	 * Build the access chain to reach the possibly nested canary member from
	 * the top level object type.
	 */
	while (is_canary_type(type)) {
		field = get_canary_field(type);
		type = TREE_TYPE(field);
		cref = build3_loc(loc, COMPONENT_REF, type, cref, field, NULL_TREE);
	}

	return cref;
}

static tree canary_value(tree field)
{
	tree value;

	gcc_assert(TREE_CODE(field) == FIELD_DECL);

	value = lookup_attribute("canary", DECL_ATTRIBUTES(field));
	gcc_assert(value);

	/* TODO: Might be a list in the future! */
	value = TREE_VALUE(TREE_VALUE(value));
	if (!useless_type_conversion_p(TREE_TYPE(field), TREE_TYPE(value))) {
		gcc_assert(TREE_CODE(value) == INTEGER_CST);
		value = fold_convert(TREE_TYPE(field), value);
	}

	return value;
}

static void canary_add_check(gimple_stmt_iterator *gsi, tree ref)
{
	tree canary, magic_val, tmp_var;
	gimple_stmt_ptr stmt, canary_check;
	basic_block cond_bb, join_bb, abort_bb;
	location_t loc;
	tree block;
	edge e;

	stmt = gsi_stmt(*gsi);
	loc = EXPR_LOC_OR_LOC(ref, gimple_location(stmt));
	block = gimple_block(stmt);

	if (verbose)
		inform(loc, "instrumenting %<canary%> type'd read%s", LOCATION_LOCUS(loc) != UNKNOWN_LOCATION ? " at:" : "");

//	fprintf(stderr, "%s: stmt: ", __func__); debug_gimple_stmt(stmt);

	/*
	 * gsi / stmt points to an instruction dereferencing a canary typed object.
	 * Insert a canary check in front of it:
	 *
	 * > typeof(var->canary) tmp = var->canary;
	 * > if (unlikely(tmp != CANARY_MAGIC))
	 * > 	__buildin_trap();
	 */
	canary = canary_from_ref(ref, loc);

	gcc_assert(TREE_CODE(canary) == COMPONENT_REF);
	gcc_assert(TREE_CODE(TREE_OPERAND(canary, 1)) == FIELD_DECL);
	gcc_assert(TREE_CODE(TREE_TYPE(canary)) == INTEGER_TYPE);

	tmp_var = create_tmp_var(TREE_TYPE(canary), "canary_check");

	magic_val = canary_value(TREE_OPERAND(canary, 1));
	gcc_assert(TREE_CODE(TREE_TYPE(magic_val)) == INTEGER_TYPE);

//	fprintf(stderr, "%s: mem_ref = ", __func__); debug_tree(ref);
//	fprintf(stderr, "%s: canary = ", __func__); debug_tree(canary);
//	fprintf(stderr, "%s: tmp_var = ", __func__); debug_tree(tmp_var);

	canary_check = gimple_build_assign(tmp_var, canary);
	gimple_set_location(canary_check, loc);
	gimple_set_block(canary_check, block);
	gsi_insert_before(gsi, canary_check, GSI_CONTINUE_LINKING);
	update_stmt(canary_check);
//	fprintf(stderr, "%s: canary_check(load) = ", __func__); debug_gimple_stmt(canary_check);

	canary_check = gimple_build_cond(NE_EXPR, gimple_assign_lhs(canary_check), magic_val, NULL_TREE, NULL_TREE);
	gimple_set_location(canary_check, loc);
	gimple_set_block(canary_check, block);
	gsi_insert_after(gsi, canary_check, GSI_NEW_STMT);
	update_stmt(canary_check);
//	fprintf(stderr, "%s: canary_check(cmp) = ", __func__); debug_gimple_stmt(canary_check);

	cond_bb = gimple_bb(gsi_stmt(*gsi));
	gcc_assert(!gsi_end_p(*gsi));
	gcc_assert(canary_check == gsi_stmt(*gsi));

	e = split_block(cond_bb, gsi_stmt(*gsi));
	gcc_assert(cond_bb == e->src);
	join_bb = e->dest;
	e->flags = EDGE_FALSE_VALUE;
	e->probability -= REG_BR_PROB_VERY_UNLIKELY;

	abort_bb = create_empty_bb(join_bb);
	e = make_edge(cond_bb, abort_bb, EDGE_TRUE_VALUE);
	e->probability = REG_BR_PROB_VERY_UNLIKELY;
	abort_bb->count = edge_count(e);

	gcc_assert(dom_info_available_p(CDI_DOMINATORS));
	set_immediate_dominator(CDI_DOMINATORS, abort_bb, cond_bb);
	set_immediate_dominator(CDI_DOMINATORS, join_bb, cond_bb);

	gcc_assert(cond_bb->loop_father == join_bb->loop_father);
	add_bb_to_loop(abort_bb, cond_bb->loop_father);
	loops_state_set(LOOPS_NEED_FIXUP);

	*gsi = gsi_start_bb(abort_bb);

	canary_check = gimple_build_call(builtin_decl_implicit(BUILT_IN_TRAP), 0);
	gimple_set_location(canary_check, loc);
	gimple_set_block(canary_check, block);
	gsi_insert_after(gsi, canary_check, GSI_CONTINUE_LINKING);
//	fprintf(stderr, "%s: canary_check(trap) = ", __func__); debug_gimple_stmt(canary_check);

	*gsi = gsi_start_nondebug_after_labels_bb(join_bb);
}

static tree find_uninstrumented_ptr(tree *tp, int *walk_subtrees, void *data)
{
	tree res = NULL_TREE;
	tree op = *tp;

	/* Look for MEM_REFs to canary'ed types. */
	switch (TREE_CODE(op)) {
	default:
		error_at(EXPR_LOCATION(op), "canary:%s(): unhandled tree type %qs", __func__, get_tree_code_name(TREE_CODE(op)));
		debug_tree(op);
		break;

	case MEM_REF:
	case TARGET_MEM_REF: {
		tree arg, type;

		if (is_canary_type(TREE_TYPE(op))) {
			res = op;
			break;
		}

		/*
		 * Look for a possibly offset MEM_REF using a narrowed type but based
		 * on an SSA_NAME that needs instrumentation.
		 */

		arg = TREE_OPERAND(op, 0);
		if (TREE_CODE(arg) != SSA_NAME)
			break;

		type = TREE_TYPE(arg);
		if (!POINTER_TYPE_P(type))
			break;

		if (!is_canary_type(TREE_TYPE(type)))
			break;

		/* A narrowed MEM_REF -- will handle the upcast in canary_from_ref(). */
//		inform(EXPR_LOCATION(op), "narrowed %<canary%> MEM_REF found");
//		fprintf(stderr, "op: "); debug_tree(op);
//		fprintf(stderr, "type: "); debug_tree(type);
		res = op;
		break;
	}

	case ARRAY_REF:
	case BIT_FIELD_REF:
	case COMPONENT_REF:
	case POINTER_PLUS_EXPR:
	case VIEW_CONVERT_EXPR:
		/* nodes we need to dig in -- the first operand only, though */
		break;

	case ADDR_EXPR:
		/*
		 * Taking the address of an object only requires instrumentation if
		 * that address gets dereferenced for reading later on. Within the
		 * same function we'd see those as MEM_REFs, which will be handled.
		 * The only two exceptions would be call arguments passed to other
		 * functions and asm statements.
		 *
		 * We could handle the latter but these would likely only catch
		 * atomic ops and there's no real point to, as we care about "real"
		 * uses of members -- not the refcounting of the object.
		 *
		 * Handling call arguments would need an IPA pass to know if the
		 * passed address gets read or written to. So far we have no need
		 * for it for the object types handled by the 'canary' attribute.
		 */
		/* fall-through */
	case CONSTRUCTOR:
	case EQ_EXPR:
	case NE_EXPR:
	case GE_EXPR:
	case GT_EXPR:
	case LE_EXPR:
	case LT_EXPR:
	case VAR_DECL:
	case PARM_DECL:
	case RESULT_DECL:
	case FUNCTION_DECL:
	case FIELD_DECL:
	case LABEL_DECL:
	case CASE_LABEL_EXPR:
	case INTEGER_CST:
	case STRING_CST:
	case VECTOR_CST:
#if BUILDING_GCC_VERSION > 8000
	case POLY_INT_CST:
#endif
	case FIXED_CST:
	case REAL_CST:
	case SSA_NAME:
	case REALPART_EXPR:	// (ab)used by __builtin_*_overflow()
	case IMAGPART_EXPR:	//          -- " --
	case COMPLEX_CST:	//          -- " --
		/* uninteresting trees */
		*walk_subtrees = 0;
		break;
	}

	return res;
}

static bool canary_handle_stmt(gimple_stmt_iterator *gsi)
{
	gimple_stmt_ptr stmt = gsi_stmt(*gsi);
	bool changed = false;
	int i = -1;

	switch (gimple_code(stmt)) {
	default:
		error_at(gimple_location(stmt), "canary:%s(): unhandled statement %qs", __func__, gimple_code_name[gimple_code(stmt)]);
		debug_gimple_stmt(stmt);
		gcc_unreachable();
		break;

	case GIMPLE_ASM:
		/* XXX: Assume inline asm behaves -- as in only be used for atomic ops. */
		break;

	case GIMPLE_ASSIGN:
		i = 1;	// rhs operand(s)
		break;

	case GIMPLE_CALL:
		i = 3;	// call arguments
		break;

	case GIMPLE_RETURN:
		i = 0;	// return value
		break;
	}

	if (i >= 0) {
		tree ptr, *op_ptr;

		/* Walk the operands and instrument where necessary. */
		for (; i < gimple_num_ops(stmt); i++) {
			op_ptr = gimple_op_ptr(stmt, i);
			ptr = walk_tree_without_duplicates(op_ptr, find_uninstrumented_ptr, NULL);
			if (ptr != NULL_TREE) {
				canary_add_check(gsi, ptr);
				changed = true;
			}
		}
	}

	return changed;
}

/*
 * This pass looks for statements involving memory references that are part of
 * or are based on a '*has_canary' marked type (which was propagated from the
 * initial in-source 'canary' attribute marking). If such a reference gets
 * dereferenced for a read, it gets instrumented with a canary check.
 *
 * We don't try to optimize multiple reads of the same memory reference and
 * rely on later passes to DCE unneeded ones. So far, this is working quite
 * well, as later passes are able to reason which statements may modify memory
 * and even move canary checks out of loops.
 *
 * Considering the following function and assuming that 'struct cred' has a
 * canary member:
 *
 *  10 bool test_creds(struct cred *cred)
 *  11 {
 *  12     const struct cred *new, *old;
 *  13     kuid_t uid;
 *  14     bool res;
 *  15
 *  16     new = prepare_kernel_cred(NULL);
 *  17     old = override_creds(new);
 *  18     revert_creds(old);
 *  18
 *  20     if (userns_may_setgroups(cred->user_ns))
 *  21         return true;
 *  22
 *  23     uid = cred->uid;
 *  24     res = uid_eq(uid, GLOBAL_ROOT_UID) && !(cred->securebits & SECBIT_NOROOT);
 *  25
 *  26     return res;
 *  27 }
 *
 * None of the 'cred' objects in lines 16-18 will be instrumented, as they do
 * not get dereferenced in this function -- only pointers get passed around.
 * If they are dereferenced in either override_creds() or revert_creds(), those
 * accesses will be instrumented in the corresponding function instead.
 *
 * Lines 20, 23 and 24 dereference the 'cred' object ('->user_ns' for the call
 * argument, '->uid' for the variable assignment and '->securebits' for the bit
 * test, resp.). All of these get instrumented, as can be seen in the following
 * gimple dump of the pass (slightly stripped down for readability):
 *
 * 100 test_creds (struct cred * cred)
 * 101 {
 * 102   unsigned int canary_check.2905;
 * 102   unsigned int canary_check.2904;
 * 104   unsigned int canary_check.2903;
 * 105   unsigned int uid$val;
 * 106   bool res;
 * 107   const struct cred * old;
 * 108   const struct cred * new;
 * 109   struct user_namespace * _1;
 * 110   _Bool _2;
 * 111   unsigned int _3;
 * 112   unsigned int _4;
 * 113   bool _5;
 * 114   int iftmp.11_6;
 * 115
 * 116   <bb 2> [local count: 1073741824]:
 * 117   new_9 = prepare_kernel_cred (0B);
 * 118   old_11 = override_creds (new_9);
 * 119   revert_creds (old_11);
 * 120   canary_check.2903_18 = cred_13(D)->canary;
 * 121   if (canary_check.2903_18 != 3119657295)
 * 122     goto <bb 9>; [0.05%]
 * 123   else
 * 124     goto <bb 8>; [99.95%]
 * 125
 * 126   <bb 8> [local count: 1073741824]:
 * 127   _1 = cred_13(D)->user_ns;
 * 128   _2 = userns_may_setgroups (_1);
 * 129   if (_2 != 0)
 * 130     goto <bb 7>; [34.00%]
 * 131   else
 * 132     goto <bb 3>; [66.00%]
 * 133
 * 134   <bb 9> [local count: 536864]:
 * 135   __builtin_trap ();
 * 136
 * 137   <bb 3> [local count: 708669601]:
 * 138   canary_check.2904_20 = cred_13(D)->canary;
 * 139   if (canary_check.2904_20 != 3119657295)
 * 140     goto <bb 11>; [0.05%]
 * 141   else
 * 142     goto <bb 10>; [99.95%]
 * 143
 * 144   <bb 10> [local count: 708669601]:
 * 145   uid$val_17 = cred_13(D)->uid.val;
 * 146   if (uid$val_17 == 0)
 * 147     goto <bb 4>; [50.00%]
 * 148   else
 * 149     goto <bb 5>; [50.00%]
 * 150
 * 151   <bb 11> [local count: 354330]:
 * 152   __builtin_trap ();
 * 153
 * 154   <bb 4> [local count: 354334800]:
 * 155   canary_check.2905_22 = cred_13(D)->canary;
 * 156   if (canary_check.2905_22 != 3119657295)
 * 157     goto <bb 13>; [0.05%]
 * 158   else
 * 159     goto <bb 12>; [99.95%]
 * 160
 * 161   <bb 12> [local count: 354334800]:
 * 162   _3 = cred_13(D)->securebits;
 * 163   _4 = _3 & 1;
 * 164   if (_4 == 0)
 * 165     goto <bb 6>; [50.00%]
 * 166   else
 * 167     goto <bb 5>; [50.00%]
 * 168
 * 169   <bb 13> [local count: 177165]:
 * 170   __builtin_trap ();
 * 171
 * 172   <bb 5> [local count: 531502200]:
 * 173
 * 174   <bb 6> [local count: 708669601]:
 * 175   # iftmp.11_6 = PHI <1(12), 0(5)>
 * 176   res_16 = (bool) iftmp.11_6;
 * 177
 * 178   <bb 7> [local count: 1073741824]:
 * 179   # _5 = PHI <1(8), res_16(6)>
 * 180   return _5;
 * 181 }
 *
 * Lines 120-124 are the canary check for the '->user_ns' dereference of line
 * 20 in the source code. The canary value is a compile time random value and
 * is 3119657295 in the above dump. Lines 137-142 is the check for the
 * '->uid.val' load of the already inlined and optimized call to uid_eq(). It's
 * followed by yet another canary check in lines 154-159 for the '->securebits'
 * load.
 *
 * Even though the last check seems superfluous as there is no memory write
 * operation between the '->uid' and '->securebits' loads that could modify the
 * 'cred' object, there's also no need to do sophisticated data flow analysis.
 * Unneeded checks will be optimized away by later passes, as can be seen in
 * the below gimple dump of the 'optimized' pass:
 *
 * 100 test_creds (struct cred * cred)
 * 101 {
 * 102   unsigned int canary_check.2904;
 * 102   unsigned int canary_check.2903;
 * 104   unsigned int uid$val;
 * 105   const struct cred * old;
 * 106   const struct cred * new;
 * 107   struct user_namespace * _1;
 * 108   _Bool _2;
 * 109   unsigned int _3;
 * 110   unsigned int _4;
 * 111   bool _5;
 * 112   unsigned int _6;
 * 113   _Bool _21;
 * 114   _Bool prephitmp_22;
 * 115
 * 116   <bb 2> [local count: 1073741824]:
 * 117   new_9 = prepare_kernel_cred (0B);
 * 118   old_11 = override_creds (new_9);
 * 119   revert_creds (old_11);
 * 120   canary_check.2903_18 = cred_13(D)->canary;
 * 121   if (canary_check.2903_18 != 3119657295)
 * 122     goto <bb 4>; [0.05%]
 * 123   else
 * 124     goto <bb 3>; [99.95%]
 * 125
 * 126   <bb 3> [local count: 1073741824]:
 * 127   _1 = cred_13(D)->user_ns;
 * 128   _2 = userns_may_setgroups (_1);
 * 129   if (_2 != 0)
 * 130     goto <bb 9>; [34.00%]
 * 131   else
 * 132     goto <bb 5>; [66.00%]
 * 133
 * 134   <bb 4> [local count: 891194]:
 * 135   __builtin_trap ();
 * 136
 * 137   <bb 5> [local count: 708669601]:
 * 138   canary_check.2904_20 = cred_13(D)->canary;
 * 139   if (canary_check.2904_20 != 3119657295)
 * 140     goto <bb 4>; [0.05%]
 * 141   else
 * 142     goto <bb 6>; [99.95%]
 * 143
 * 144   <bb 6> [local count: 708669601]:
 * 145   uid$val_17 = cred_13(D)->uid.val;
 * 146   if (uid$val_17 == 0)
 * 147     goto <bb 7>; [50.00%]
 * 148   else
 * 149     goto <bb 8>; [50.00%]
 * 150
 * 151   <bb 7> [local count: 354334800]:
 * 152   _3 = cred_13(D)->securebits;
 * 153   _4 = _3 & 1;
 * 154   _6 = 1 - _4;
 * 155   _21 = (_Bool) _6;
 * 156
 * 157   <bb 8> [local count: 708669601]:
 * 158   # prephitmp_22 = PHI <_21(7), _2(6)>
 * 159
 * 160   <bb 9> [local count: 1073741824]:
 * 161   # _5 = PHI <prephitmp_22(8), _2(3)>
 * 162   return _5;
 * 163 }
 *
 * Only two checks remain: the initial one for the '->user_ns' dereference
 * and that one preceeding the '->uid' and '->securebits' loads.
 */
static unsigned int type_canary_execute(void)
{
	basic_block bb;

	loop_optimizer_init(LOOPS_NORMAL | LOOPS_HAVE_RECORDED_EXITS);
	gcc_assert(current_loops);

	calculate_dominance_info(CDI_DOMINATORS);
	calculate_dominance_info(CDI_POST_DOMINATORS);

	/* Evaluate all nodes and instrument, where needed. */
	FOR_EACH_BB_FN(bb, cfun) {
		gimple_stmt_iterator gsi;

		for (gsi = gsi_start_bb(bb); !gsi_end_p(gsi); gsi_next(&gsi)) {
			gimple_stmt_ptr stmt = gsi_stmt(gsi);

			if (!gimple_references_memory_p(stmt))
				continue;

			if (canary_handle_stmt(&gsi)) {
				/*
				 * Instrumenting a statement hijacks its BB requiring us to
				 * update the 'bb' pointer to avoid seeing (and instrumenting)
				 * statements twice. We will, however, see the BB for the call
				 * to __builtin_trap(). But that's benign, as neither will it
				 * match our search pattern nor is it big enough to cause any
				 * measurable compile time slowdown.
				 */
				bb = gsi_bb(gsi);
			}

			gcc_assert(gsi_stmt(gsi) == stmt);
		}
	}

	free_dominance_info(CDI_DOMINATORS);
	free_dominance_info(CDI_POST_DOMINATORS);
	loop_optimizer_finalize();

	return 0;
}

static bool type_canary_gate(void)
{
	if (lookup_attribute("no_instrument_function", DECL_ATTRIBUTES(current_function_decl)))
		return false;

	if (has_trait(current_function_decl, "nocanary"))
		return false;

	return true;
}

#define PASS_NAME type_canary
#define PROPERTIES_REQUIRED PROP_cfg
#define TODO_FLAGS_FINISH TODO_verify_il | TODO_verify_ssa | TODO_verify_stmts | TODO_dump_func | TODO_update_ssa | TODO_verify_flow | TODO_update_address_taken

#include "gcc-generate-gimple-pass.h"

__visible int plugin_init(struct plugin_name_args *plugin_info, struct plugin_gcc_version *version)
{
	const char *const plugin_name = plugin_info->base_name;
	const struct plugin_argument *argv = plugin_info->argv;
	const int argc = plugin_info->argc;
	int i;

	PASS_INFO(type_canary, "sra", 0, PASS_POS_INSERT_AFTER);

	if (!plugin_default_version_check(version, &gcc_version)) {
		error_gcc_version(version);
		return 1;
	}

	register_callback(plugin_name, PLUGIN_INFO, NULL, &type_canary_plugin_info);

	for (i = 0; i < argc; ++i) {
		if (!strcmp(argv[i].key, "verbose")) {
			verbose++;
			continue;
		}

		error(G_("unknown option '-fplugin-arg-%s-%s'"), plugin_name, argv[i].key);
	}

	register_callback(plugin_name, PLUGIN_ATTRIBUTES, register_attributes, NULL);
	register_callback(plugin_name, PLUGIN_FINISH_TYPE, finish_type, NULL);
	register_callback(plugin_name, PLUGIN_ALL_PASSES_START, typemap_cleanup, NULL);
	register_callback(plugin_name, PLUGIN_PASS_MANAGER_SETUP, NULL, &type_canary_pass_info);

	return 0;
}
